# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('openstack', '0024_network_subnet'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='backup',
            name='backup_schedule',
        ),
        migrations.RemoveField(
            model_name='backup',
            name='instance',
        ),
        migrations.RemoveField(
            model_name='backup',
            name='snapshots',
        ),
        migrations.RemoveField(
            model_name='backup',
            name='tenant',
        ),
        migrations.RemoveField(
            model_name='backuprestoration',
            name='backup',
        ),
        migrations.RemoveField(
            model_name='backuprestoration',
            name='flavor',
        ),
        migrations.RemoveField(
            model_name='backuprestoration',
            name='instance',
        ),
        migrations.RemoveField(
            model_name='backupschedule',
            name='instance',
        ),
        migrations.RemoveField(
            model_name='drbackup',
            name='backup_schedule',
        ),
        migrations.RemoveField(
            model_name='drbackup',
            name='service_project_link',
        ),
        migrations.RemoveField(
            model_name='drbackup',
            name='source_instance',
        ),
        migrations.RemoveField(
            model_name='drbackup',
            name='tags',
        ),
        migrations.RemoveField(
            model_name='drbackup',
            name='temporary_snapshots',
        ),
        migrations.RemoveField(
            model_name='drbackup',
            name='temporary_volumes',
        ),
        migrations.RemoveField(
            model_name='drbackup',
            name='tenant',
        ),
        migrations.RemoveField(
            model_name='drbackup',
            name='volume_backups',
        ),
        migrations.RemoveField(
            model_name='drbackuprestoration',
            name='backup',
        ),
        migrations.RemoveField(
            model_name='drbackuprestoration',
            name='flavor',
        ),
        migrations.RemoveField(
            model_name='drbackuprestoration',
            name='instance',
        ),
        migrations.RemoveField(
            model_name='drbackuprestoration',
            name='tenant',
        ),
        migrations.RemoveField(
            model_name='drbackuprestoration',
            name='volume_backup_restorations',
        ),
        migrations.RemoveField(
            model_name='instance',
            name='security_groups',
        ),
        migrations.RemoveField(
            model_name='instance',
            name='service_project_link',
        ),
        migrations.RemoveField(
            model_name='instance',
            name='tags',
        ),
        migrations.RemoveField(
            model_name='instance',
            name='tenant',
        ),
        migrations.RemoveField(
            model_name='snapshot',
            name='service_project_link',
        ),
        migrations.RemoveField(
            model_name='snapshot',
            name='source_volume',
        ),
        migrations.RemoveField(
            model_name='snapshot',
            name='tags',
        ),
        migrations.RemoveField(
            model_name='snapshot',
            name='tenant',
        ),
        migrations.RemoveField(
            model_name='volume',
            name='image',
        ),
        migrations.RemoveField(
            model_name='volume',
            name='instance',
        ),
        migrations.RemoveField(
            model_name='volume',
            name='service_project_link',
        ),
        migrations.RemoveField(
            model_name='volume',
            name='source_snapshot',
        ),
        migrations.RemoveField(
            model_name='volume',
            name='tags',
        ),
        migrations.RemoveField(
            model_name='volume',
            name='tenant',
        ),
        migrations.RemoveField(
            model_name='volumebackup',
            name='record',
        ),
        migrations.RemoveField(
            model_name='volumebackup',
            name='service_project_link',
        ),
        migrations.RemoveField(
            model_name='volumebackup',
            name='source_volume',
        ),
        migrations.RemoveField(
            model_name='volumebackup',
            name='tags',
        ),
        migrations.RemoveField(
            model_name='volumebackup',
            name='tenant',
        ),
        migrations.RemoveField(
            model_name='volumebackuprestoration',
            name='mirorred_volume_backup',
        ),
        migrations.RemoveField(
            model_name='volumebackuprestoration',
            name='tenant',
        ),
        migrations.RemoveField(
            model_name='volumebackuprestoration',
            name='volume',
        ),
        migrations.RemoveField(
            model_name='volumebackuprestoration',
            name='volume_backup',
        ),
        migrations.DeleteModel(
            name='Backup',
        ),
        migrations.DeleteModel(
            name='BackupRestoration',
        ),
        migrations.DeleteModel(
            name='BackupSchedule',
        ),
        migrations.DeleteModel(
            name='DRBackup',
        ),
        migrations.DeleteModel(
            name='DRBackupRestoration',
        ),
        migrations.DeleteModel(
            name='Instance',
        ),
        migrations.DeleteModel(
            name='Snapshot',
        ),
        migrations.DeleteModel(
            name='Volume',
        ),
        migrations.DeleteModel(
            name='VolumeBackup',
        ),
        migrations.DeleteModel(
            name='VolumeBackupRecord',
        ),
        migrations.DeleteModel(
            name='VolumeBackupRestoration',
        ),
    ]



import json
import logging
import os
import subprocess  # nosec

import six
from django.conf import settings

from waldur_ansible.common import exceptions
from waldur_core.core.views import RefreshTokenMixin

from . import utils

logger = logging.getLogger(__name__)


class ManagementRequestsBackend(object):

    def is_processing_allowed(self, request):
        raise NotImplementedError

    def build_locked_for_processing_message(self, request):
        raise NotImplementedError

    def lock_for_processing(self, request):
        raise NotImplementedError

    def handle_on_processing_finished(self, request):
        raise NotImplementedError

    def get_playbook_path(self, request):
        raise NotImplementedError

    def get_user(self, request):
        raise NotImplementedError

    def build_additional_extra_vars(self, request):
        raise NotImplementedError

    def instantiate_line_post_processor_class(self, request):
        raise NotImplementedError

    def instantiate_extracted_information_handler_class(self, request):
        raise NotImplementedError

    def instantiate_error_handler_class(self, request):
        raise NotImplementedError

    def process_request(self, request):
        if not self.is_processing_allowed(request):
            request.output = self.build_locked_for_processing_message(request)
            request.save(update_fields=['output'])
            raise exceptions.LockedForProcessingError('Could not process request %s ' % request)
        try:
            self.lock_for_processing(request)

            command = self.build_command(request)
            command_str = ' '.join(command)

            logger.debug('Executing command "%s".', command_str)
            env = dict(
                os.environ,
                ANSIBLE_LIBRARY=settings.WALDUR_ANSIBLE_COMMON['ANSIBLE_LIBRARY'],
                ANSIBLE_HOST_KEY_CHECKING='False',
                ANSIBLE_RETRY_FILES_ENABLED='False',
                ANSIBLE_REMOTE_PORT=settings.WALDUR_ANSIBLE_COMMON['REMOTE_VM_SSH_PORT'],
            )
            lines_post_processor_instance = self.instantiate_line_post_processor_class(request)
            extracted_information_handler = self.instantiate_extracted_information_handler_class(request)
            error_handler = self.instantiate_error_handler_class(request)
            try:
                for output_line in utils.subprocess_output_iterator(command, env):
                    request.output += output_line
                    request.save(update_fields=['output'])
                    lines_post_processor_instance.post_process_line(output_line)
            except subprocess.CalledProcessError as e:
                logger.error('%s - failed to execute command "%s".', request, command_str)
                logger.error('%s - Ansible request processing output: \n %s.', request, request.output)
                error_handler.handle_error(request, lines_post_processor_instance)
                six.reraise(exceptions.AnsibleBackendError, e)
            else:
                logger.info('Command "%s" was successfully executed.', command_str)
                extracted_information_handler.handle_extracted_information(request, lines_post_processor_instance)
        finally:
            self.handle_on_processing_finished(request)

    def build_command(self, request):
        playbook_path = self.get_playbook_path(request)
        self.ensure_playbook_exists_or_raise(playbook_path)

        command = [settings.WALDUR_ANSIBLE_COMMON.get('PLAYBOOK_EXECUTION_COMMAND', 'ansible-playbook')]

        if settings.WALDUR_ANSIBLE_COMMON.get('PLAYBOOK_ARGUMENTS'):
            command.extend(settings.WALDUR_ANSIBLE_COMMON.get('PLAYBOOK_ARGUMENTS'))

        command.extend(['--extra-vars', self.build_extra_vars(request)])

        command.extend(['--ssh-common-args', '-o UserKnownHostsFile=/dev/null'])

        return command + [playbook_path]

    def ensure_playbook_exists_or_raise(self, playbook_path):
        if not os.path.exists(playbook_path):
            raise exceptions.AnsibleBackendError('Playbook %s does not exist.' % playbook_path)

    def build_extra_vars(self, request):
        extra_vars = self.build_common_extra_vars(request)
        extra_vars.update(self.build_additional_extra_vars(request))
        return json.dumps(extra_vars)

    def build_common_extra_vars(self, request):
        return dict(
            api_url=settings.WALDUR_ANSIBLE_COMMON['API_URL'],
            access_token=RefreshTokenMixin().refresh_token(self.get_user(request)).key,
            private_key_path=settings.WALDUR_ANSIBLE_COMMON['PRIVATE_KEY_PATH'],
            public_key_uuid=settings.WALDUR_ANSIBLE_COMMON['PUBLIC_KEY_UUID'],
        )



from collections import defaultdict

from rest_framework import status, exceptions, response
from rest_framework.decorators import detail_route, list_route
from rest_framework.generics import get_object_or_404
from rest_framework.response import Response

from waldur_core.core.exceptions import IncorrectStateException
from waldur_core.core.serializers import HistorySerializer
from waldur_core.core.utils import datetime_to_timestamp, pwgen
from waldur_core.monitoring.utils import get_period
from waldur_core.structure import views as structure_views

from . import models, serializers, filters, executors
from .managers import filter_active


class ZabbixServiceViewSet(structure_views.BaseServiceViewSet):
    queryset = models.ZabbixService.objects.all()
    serializer_class = serializers.ServiceSerializer

    def get_serializer_class(self):
        if self.action == 'credentials':
            return serializers.UserSerializer
        elif self.action == 'trigger_status':
            return serializers.TriggerRequestSerializer
        return super(ZabbixServiceViewSet, self).get_serializer_class()

    @detail_route(methods=['GET', 'POST'])
    def credentials(self, request, uuid):
        """ On GET request - return superadmin user data.
            On POST - reset superuser password and return new one.
        """
        service = self.get_object()
        if request.method == 'GET':
            user = models.User.objects.get(settings=service.settings, alias=service.settings.username)
            serializer_class = self.get_serializer_class()
            serializer = serializer_class(user, context=self.get_serializer_context())
            return Response(serializer.data)
        else:
            password = pwgen()
            executors.ServiceSettingsPasswordResetExecutor.execute(service.settings, password=password)
            return Response({'password': password})

    @detail_route(methods=['GET', 'HEAD'])
    def trigger_status(self, request, uuid):
        serializer_class = self.get_serializer_class()
        serializer = serializer_class(data=request.query_params)
        serializer.is_valid(raise_exception=True)
        query = serializer.validated_data

        service = self.get_object()
        backend = service.get_backend()
        headers = {
            'X-Result-Count': backend.get_trigger_count(query),
        }

        if request.method == 'HEAD':
            return Response(headers=headers)

        backend_triggers = backend.get_trigger_status(query)
        response_serializer = serializers.TriggerResponseSerializer(
            instance=backend_triggers, many=True)

        page = self.paginate_queryset(response_serializer.data)
        if page is not None:
            return self.get_paginated_response(page)

        return Response(response_serializer.data, headers=headers)


class ZabbixServiceProjectLinkViewSet(structure_views.BaseServiceProjectLinkViewSet):
    queryset = models.ZabbixServiceProjectLink.objects.all()
    serializer_class = serializers.ServiceProjectLinkSerializer


class NoHostsException(exceptions.APIException):
    status_code = status.HTTP_400_BAD_REQUEST
    default_detail = 'There are no OK hosts that match given query.'


class NoItemsException(exceptions.APIException):
    status_code = status.HTTP_400_BAD_REQUEST
    default_detail = 'There are no items that match given query.'


class HostViewSet(structure_views.ResourceViewSet):
    """ Representation of Zabbix hosts and related actions. """
    queryset = models.Host.objects.all()
    serializer_class = serializers.HostSerializer
    filter_backends = structure_views.ResourceViewSet.filter_backends + (
        filters.HostScopeFilterBackend,
    )
    create_executor = executors.HostCreateExecutor
    update_executor = executors.HostUpdateExecutor
    delete_executor = executors.HostDeleteExecutor

    @detail_route()
    def items_history(self, request, uuid):
        """ Get host items historical values.

        Request should specify datetime points and items.
        There are two ways to define datetime points for historical data.

        1. Send *?point=<timestamp>* parameter that can list.
           Response will contain historical data for each given point in the same order.
        2. Send *?start=<timestamp>*, *?end=<timestamp>*, *?points_count=<integer>* parameters.
           Result will contain <points_count> points from <start> to <end>.

        Also you should specify one or more name of host template items, for example 'openstack.instance.cpu_util'

        Response is list of datapoints, each of which is dictionary with following fields:
         - 'point' - timestamp;
         - 'value' - values are converted from bytes to megabytes, if possible;
         - 'item' - key of host template item;
         - 'item_name' - name of host template item.
        """
        host = self.get_object()
        if host.state != models.Host.States.OK:
            raise IncorrectStateException('Host has to be OK to get items history.')
        stats = self._get_stats(request, [host])
        return Response(stats, status=status.HTTP_200_OK)

    @list_route()
    def aggregated_items_history(self, request):
        """ Get sum of hosts historical values.

        Request should specify host filtering parameters, datetime points, and items.
        Host filtering parameters are the same as for */api/zabbix-hosts/* endpoint.
        Input/output format is the same as for **/api/zabbix-hosts/<host_uuid>/items_history/** endpoint.
        """
        stats = self._get_stats(request, self._get_hosts())
        return Response(stats, status=status.HTTP_200_OK)

    @list_route()
    def items_aggregated_values(self, request):
        """ Get sum of aggregated hosts values.

        Request parameters:
         - ?start - start of aggregation period as timestamp. Default: 1 hour ago.
         - ?end - end of aggregation period as timestamp. Default: now.
         - ?method - aggregation method. Default: MAX. Choices: MIN, MAX.
         - ?item - item key. Can be list. Required.

        Response format: {<item key>: <aggregated value>, ...}

        Endpoint will return status 400 if there are no hosts or items that match request parameters.
        """
        hosts = self._get_hosts()
        serializer = serializers.ItemsAggregatedValuesSerializer(data=request.query_params)
        serializer.is_valid(raise_exception=True)
        filter_data = serializer.validated_data
        items = self._get_items(request, hosts)

        aggregated_data = defaultdict(lambda: 0)
        for host in hosts:
            backend = host.get_backend()
            host_aggregated_values = backend.get_items_aggregated_values(
                host, items, filter_data['start'], filter_data['end'], filter_data['method'])
            for key, value in host_aggregated_values.items():
                aggregated_data[key] += value
        return Response(aggregated_data, status=status.HTTP_200_OK)

    # TODO: make methods items_aggregated_values and items_values DRY.
    @detail_route()
    def items_values(self, request, uuid):
        """ The same as items_aggregated_values, only for one host """
        host = self.get_object()
        serializer = serializers.ItemsAggregatedValuesSerializer(data=request.query_params)
        serializer.is_valid(raise_exception=True)
        filter_data = serializer.validated_data
        items = self._get_items(request, [host])

        backend = host.get_backend()
        host_aggregated_values = backend.get_items_aggregated_values(
            host, items, filter_data['start'], filter_data['end'], filter_data['method'])
        return Response(host_aggregated_values, status=status.HTTP_200_OK)

    def _get_hosts(self):
        hosts = filter_active(self.filter_queryset(self.get_queryset()))
        if not hosts:
            raise NoHostsException()
        return hosts

    def _get_items(self, request, hosts):
        items = request.query_params.getlist('item')
        items = models.Item.objects.filter(template__hosts__in=hosts, key__in=items).distinct()
        if not items:
            raise NoItemsException()
        return items

    def _get_stats(self, request, hosts):
        """
        If item list contains several elements, result is ordered by item
        (in the same order as it has been provided in request) and then by time.
        """
        items = self._get_items(request, hosts)
        numeric_types = (models.Item.ValueTypes.FLOAT, models.Item.ValueTypes.INTEGER)
        non_numeric_items = [item.name for item in items if item.value_type not in numeric_types]
        if non_numeric_items:
            raise exceptions.ValidationError(
                'Cannot show historical data for non-numeric items: %s' % ', '.join(non_numeric_items))
        points = self._get_points(request)

        stats = []
        for item in items:
            values = self._sum_rows([
                host.get_backend().get_item_stats(host.backend_id, item, points)
                for host in hosts
            ])

            for point, value in zip(points, values):
                stats.append({
                    'point': point,
                    'item': item.key,
                    'item_name': item.name,
                    'value': value,
                })
        return stats

    def _get_points(self, request):
        mapped = {
            'start': request.query_params.get('start'),
            'end': request.query_params.get('end'),
            'points_count': request.query_params.get('points_count'),
            'point_list': request.query_params.getlist('point')
        }
        serializer = HistorySerializer(data={k: v for k, v in mapped.items() if v})
        serializer.is_valid(raise_exception=True)
        points = map(datetime_to_timestamp, serializer.get_filter_data())
        return points

    def _sum_rows(self, rows):
        """
        Input: [[1, 2], [10, 20], [None, None]]
        Output: [11, 22]
        """
        def sum_without_none(xs):
            return sum(x for x in xs if x)
        return map(sum_without_none, zip(*rows))


class ITServiceViewSet(structure_views.ResourceViewSet):
    queryset = models.ITService.objects.all().select_related('trigger')
    serializer_class = serializers.ITServiceSerializer
    lookup_field = 'uuid'
    create_executor = executors.ITServiceCreateExecutor
    delete_executor = executors.ITServiceDeleteExecutor
    # TODO: add update operation

    @detail_route()
    def events(self, request, uuid):
        itservice = self.get_object()
        period = get_period(request)

        history = get_object_or_404(models.SlaHistory, itservice=itservice, period=period)
        events = list(history.events.all().order_by('-timestamp').values('timestamp', 'state'))

        serializer = serializers.SlaHistoryEventSerializer(data=events, many=True)
        serializer.is_valid(raise_exception=True)

        return Response(serializer.data, status=status.HTTP_200_OK)


class TemplateViewSet(structure_views.BaseServicePropertyViewSet):
    queryset = models.Template.objects.all().prefetch_related('items')
    serializer_class = serializers.TemplateSerializer
    lookup_field = 'uuid'
    filter_class = filters.TemplateFilter


class TriggerViewSet(structure_views.BaseServicePropertyViewSet):
    queryset = models.Trigger.objects.all()
    serializer_class = serializers.TriggerSerializer
    lookup_field = 'uuid'
    filter_class = filters.TriggerFilter


class UserGroupViewSet(structure_views.BaseServicePropertyViewSet):
    queryset = models.UserGroup.objects.all()
    serializer_class = serializers.UserGroupSerializer
    lookup_field = 'uuid'
    filter_class = filters.UserGroupFilter


class UserViewSet(structure_views.BaseServicePropertyViewSet):
    queryset = models.User.objects.all()
    serializer_class = serializers.UserSerializer
    lookup_field = 'uuid'
    filter_class = filters.UserFilter
    create_executor = executors.UserCreateExecutor
    update_executor = executors.UserUpdateExecutor
    delete_executor = executors.UserDeleteExecutor

    @detail_route(methods=['post'])
    def password(self, request, uuid):
        user = self.get_object()
        user.password = pwgen()
        user.save()
        executors.UserUpdateExecutor.execute(user, updated_fields=['password'])
        return response.Response(
            {'detail': 'password update was scheduled successfully', 'password': user.password},
            status=status.HTTP_200_OK
        )



from functools import reduce
import logging
import operator

from django.conf import settings as django_settings
from django.db import transaction
from django.utils import timezone
import six

from waldur_core.structure import ServiceBackend, ServiceBackendError
from waldur_freeipa import models as freeipa_models
from waldur_slurm.client import SlurmClient
from waldur_slurm.client_moab import MoabClient
from waldur_slurm.structures import Quotas

from . import models, base

logger = logging.getLogger(__name__)


class SlurmBackend(ServiceBackend):
    def __init__(self, settings):
        self.settings = settings
        self.client = self.get_client(settings)

    def get_client(self, settings):
        batch_service = models.get_batch_service(settings)
        cls = SlurmClient
        if batch_service == 'MOAB':
            cls = MoabClient
        return cls(
            hostname=settings.options.get('hostname', 'localhost'),
            username=settings.username or 'root',
            port=settings.options.get('port', 22),
            key_path=django_settings.WALDUR_SLURM['PRIVATE_KEY_PATH'],
            use_sudo=settings.options.get('use_sudo', False),
        )

    def pull_resources(self):
        self.sync_usage()

    def ping(self, raise_exception=False):
        try:
            self.client.list_accounts()
        except base.BatchError as e:
            if raise_exception:
                six.reraise(ServiceBackendError, e)
            return False
        else:
            return True

    def create_allocation(self, allocation):
        project = allocation.service_project_link.project
        customer_account = self.get_customer_name(project.customer)
        project_account = self.get_project_name(project)
        allocation_account = self.get_allocation_name(allocation)

        if not self.client.get_account(customer_account):
            self.create_customer(project.customer)

        if not self.client.get_account(project_account):
            self.create_project(project)

        self.client.create_account(
            name=allocation_account,
            description=allocation.name,
            organization=project_account,
        )
        self.set_resource_limits(allocation)

        freeipa_profiles = {
            profile.user: profile.username
            for profile in freeipa_models.Profile.objects.all()
        }

        for user in allocation.service_project_link.project.customer.get_users():
            username = freeipa_profiles.get(user)
            if username:
                self.add_user(allocation, username.lower())

    def delete_allocation(self, allocation):
        account = self.get_allocation_name(allocation)
        if self.client.get_account(account):
            self.client.delete_account(account)

        project = allocation.service_project_link.project
        if self.get_allocation_queryset().filter(project=project).count() == 0:
            self.delete_project(project)

        if self.get_allocation_queryset().filter(project__customer=project.customer).count() == 0:
            self.delete_customer(project.customer)

    def add_user(self, allocation, username):
        """
        Create association between user and SLURM account if it does not exist yet.
        """
        account = self.get_allocation_name(allocation)
        default_account = self.settings.options.get('default_account')
        if not self.client.get_association(username, account):
            self.client.create_association(username, account, default_account)

    def delete_user(self, allocation, username):
        """
        Delete association between user and SLURM account if it exists.
        """
        account = self.get_allocation_name(allocation)
        if self.client.get_association(username, account):
            self.client.delete_association(username, account)

    def set_resource_limits(self, allocation):
        quotas = Quotas(
            cpu=allocation.cpu_limit,
            gpu=allocation.gpu_limit,
            ram=allocation.ram_limit,
            deposit=allocation.deposit_limit,
        )

        self.client.set_resource_limits(self.get_allocation_name(allocation), quotas)

    def cancel_allocation(self, allocation):
        allocation.cpu_limit = allocation.cpu_usage
        allocation.gpu_limit = allocation.gpu_usage
        allocation.ram_limit = allocation.ram_usage
        allocation.deposit_limit = allocation.deposit_usage

        self.set_resource_limits(allocation)

        allocation.is_active = False
        allocation.save()

    def sync_usage(self):
        waldur_allocations = {
            self.get_allocation_name(allocation): allocation
            for allocation in self.get_allocation_queryset()
        }

        report = self.get_usage_report(waldur_allocations.keys())
        for account, usage in report.items():
            allocation = waldur_allocations.get(account)
            if not allocation:
                logger.debug('Skipping usage report for account %s because it is not managed under Waldur', account)
                continue
            self._update_quotas(allocation, usage)

    def pull_allocation(self, allocation):
        account = self.get_allocation_name(allocation)
        report = self.get_usage_report([account])
        usage = report.get(account)
        if not usage:
            logger.debug('Skipping usage report for account %s because it is not managed under Waldur', account)
            return
        self._update_quotas(allocation, usage)

    def get_usage_report(self, accounts):
        report = {}
        lines = self.client.get_usage_report(accounts)

        for line in lines:
            report.setdefault(line.account, {}).setdefault(line.user, Quotas())
            report[line.account][line.user] += line.quotas

        for usage in report.values():
            quotas = usage.values()
            total = reduce(operator.add, quotas)
            usage['TOTAL_ACCOUNT_USAGE'] = total

        return report

    @transaction.atomic()
    def _update_quotas(self, allocation, usage):
        quotas = usage.pop('TOTAL_ACCOUNT_USAGE')
        allocation.cpu_usage = quotas.cpu
        allocation.gpu_usage = quotas.gpu
        allocation.ram_usage = quotas.ram
        allocation.deposit_usage = quotas.deposit
        allocation.save(update_fields=['cpu_usage', 'gpu_usage', 'ram_usage', 'deposit_usage'])

        usernames = usage.keys()
        usermap = {
            profile.username: profile.user
            for profile in freeipa_models.Profile.objects.filter(username__in=usernames)
        }

        for username, quotas in usage.items():
            models.AllocationUsage.objects.update_or_create(
                allocation=allocation,
                username=username,
                year=timezone.now().year,
                month=timezone.now().month,
                defaults={
                    'cpu_usage': quotas.cpu,
                    'gpu_usage': quotas.gpu,
                    'ram_usage': quotas.ram,
                    'deposit_usage': quotas.deposit,
                    'user': usermap.get(username),
                })

    def create_customer(self, customer):
        customer_name = self.get_customer_name(customer)
        return self.client.create_account(customer_name, customer.name, customer_name)

    def delete_customer(self, customer_uuid):
        self.client.delete_account(self.get_customer_name(customer_uuid))

    def create_project(self, project):
        name = self.get_project_name(project)
        parent_name = self.get_customer_name(project.customer)
        return self.client.create_account(name, project.name, name, parent_name)

    def delete_project(self, project_uuid):
        self.client.delete_account(self.get_project_name(project_uuid))

    def get_allocation_queryset(self):
        return models.Allocation.objects.filter(service_project_link__service__settings=self.settings)

    def get_customer_name(self, customer):
        return self.get_account_name(django_settings.WALDUR_SLURM['CUSTOMER_PREFIX'], customer)

    def get_project_name(self, project):
        return self.get_account_name(django_settings.WALDUR_SLURM['PROJECT_PREFIX'], project)

    def get_allocation_name(self, allocation):
        return self.get_account_name(django_settings.WALDUR_SLURM['ALLOCATION_PREFIX'], allocation)

    def get_account_name(self, prefix, object_or_uuid):
        key = isinstance(object_or_uuid, basestring) and object_or_uuid or object_or_uuid.uuid.hex
        return '%s%s' % (prefix, key)



import logging

from django_filters.rest_framework import DjangoFilterBackend
from rest_framework import decorators, response
from rest_framework.viewsets import GenericViewSet

from waldur_ansible.common import serializers as common_serializers
from waldur_ansible.jupyter_hub_management import models as jupyter_hub_models

from waldur_core.core import views as core_views, managers as core_managers, mixins as core_mixins
from waldur_core.structure import serializers as core_structure_serializers, filters as structure_filters
from . import models, serializers, executors, pip_service, python_management_service, utils

python_management_requests_models = [models.PythonManagementInitializeRequest,
                                     models.PythonManagementSynchronizeRequest,
                                     models.PythonManagementFindVirtualEnvsRequest,
                                     models.PythonManagementFindInstalledLibrariesRequest,
                                     models.PythonManagementDeleteVirtualEnvRequest,
                                     models.PythonManagementDeleteRequest]

logger = logging.getLogger(__name__)


class PythonManagementViewSet(core_mixins.AsyncExecutor, core_views.ActionsViewSet):
    lookup_field = 'uuid'
    queryset = models.PythonManagement.objects.all().order_by('pk')
    serializer_class = serializers.PythonManagementSerializer
    python_management_request_executor = executors.PythonManagementRequestExecutor
    service = python_management_service.PythonManagementService()

    filter_backends = (structure_filters.GenericRoleFilter, DjangoFilterBackend)

    def retrieve(self, request, *args, **kwargs):
        python_management = self.get_object()
        python_management_serializer = self.get_serializer(python_management)

        requests = core_managers.SummaryQuerySet(python_management_requests_models).filter(
            python_management=python_management).order_by("-created")
        requests_serializer = common_serializers.SummaryApplicationSerializer(
            requests, many=True, context={'select_output': False})

        return response.Response(
            {'python_management': python_management_serializer.data, 'requests': requests_serializer.data})

    @core_mixins.ensure_atomic_transaction
    def perform_create(self, serializer):
        python_management = serializer.save()

        virtual_environments = serializer.validated_data.get('virtual_environments')

        initialization_request = models.PythonManagementInitializeRequest(python_management=python_management)
        initialization_request.save()

        for virtual_environment in virtual_environments:
            libraries_to_install = []
            for library in virtual_environment['installed_libraries']:
                libraries_to_install.append(library)
            models.PythonManagementSynchronizeRequest.objects.create(
                python_management=python_management,
                initialization_request=initialization_request,
                libraries_to_install=libraries_to_install,
                virtual_env_name=virtual_environment['name'])

        self.python_management_request_executor.execute(initialization_request, async=self.async_executor)

    @core_mixins.ensure_atomic_transaction
    def perform_destroy(self, persisted_python_management):
        self.service.schedule_python_management_removal(persisted_python_management)

    @core_mixins.ensure_atomic_transaction
    def perform_update(self, serializer):
        persisted_python_management = self.get_object()
        serializer.is_valid(raise_exception=True)
        all_transient_virtual_environments = serializer.validated_data.get('virtual_environments')

        self.service.schedule_virtual_environments_update(all_transient_virtual_environments, persisted_python_management)

    @decorators.detail_route(methods=['get'])
    @core_mixins.ensure_atomic_transaction
    def find_virtual_environments(self, request, uuid=None):
        persisted_python_management = self.get_object()

        return self.service.schedule_virtual_environments_search(persisted_python_management)

    @decorators.detail_route(url_path="find_installed_libraries/(?P<virtual_env_name>.+)", methods=['get'])
    @core_mixins.ensure_atomic_transaction
    def find_installed_libraries(self, request, virtual_env_name=None, uuid=None):
        persisted_python_management = self.get_object()

        return self.service.schedule_installed_libraries_search(persisted_python_management, virtual_env_name)

    @decorators.detail_route(url_path="requests/(?P<request_uuid>.+)", methods=['get'])
    def find_request_with_output_by_uuid(self, request, uuid=None, request_uuid=None):
        requests = core_managers.SummaryQuerySet(python_management_requests_models).filter(python_management=self.get_object(),
                                                                                           uuid=request_uuid)
        serializer = common_serializers.SummaryApplicationSerializer(
            requests, many=True, context={'select_output': True})
        return response.Response(serializer.data)

    @decorators.list_route(url_path="validForJupyterHub", methods=['get'])
    def find_valid_for_jupyter_hub_python_managements_with_instance_info(self, request):
        result = super(PythonManagementViewSet, self).list(request)
        result.data = filter(
            lambda pm: not utils.execute_safely(
                lambda: jupyter_hub_models.JupyterHubManagement.objects.get(python_management__uuid=pm['uuid'])),
            result.data)

        for python_management in result.data:
            instance = models.PythonManagement.objects.get(uuid=python_management['uuid']).instance
            instance_serializer = core_structure_serializers.SummaryResourceSerializer(instance=instance, context={'request': request})
            python_management['instance'] = instance_serializer.data

        return result


class PipPackagesViewSet(GenericViewSet):

    @decorators.list_route(url_path="find_library_versions/(?P<queried_library_name>.+)/(?P<python_version>.+)", methods=['get'])
    def find_library_versions(self, request, queried_library_name=None, python_version=None):
        versions = pip_service.find_versions(queried_library_name, python_version)

        return response.Response({'versions': versions})

    @decorators.list_route(url_path="autocomplete_library/(?P<queried_library_name>.+)", methods=['get'])
    def autocomplete_library_name(self, request, queried_library_name=None):
        matching_libraries = pip_service.autocomplete_library_name(queried_library_name)
        serializer = serializers.CachedRepositoryPythonLibrarySerializer(matching_libraries, many=True)

        return response.Response(serializer.data)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('quotas', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='quota',
            name='content_type',
            field=models.ForeignKey(to='contenttypes.ContentType', null=True),
            preserve_default=True,
        ),
        migrations.AlterField(
            model_name='quota',
            name='object_id',
            field=models.PositiveIntegerField(null=True),
            preserve_default=True,
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.13 on 2018-10-08 08:19
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('openstack_tenant', '0038_internal_ip_settings_non_null'),
    ]

    operations = [
        migrations.AlterField(
            model_name='internalip',
            name='backend_id',
            field=models.CharField(max_length=255, null=True),
        ),
    ]



def build_additional_extra_args(python_management_request):
    return dict(
        virtual_env_name=python_management_request.virtual_env_name
    )


def build_sync_request_extra_args(synchronization_request):
    extra_vars = dict(
        libraries_to_install=synchronization_request.libraries_to_install,
        libraries_to_remove=synchronization_request.libraries_to_remove,
    )
    extra_vars.update(build_additional_extra_args(synchronization_request))
    return extra_vars



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models
import taggit.managers


class Migration(migrations.Migration):

    dependencies = [
        ('taggit', '0002_auto_20150616_2121'),
        ('contenttypes', '0002_remove_content_type_name'),
        ('structure', '0034_change_service_settings_state_field'),
    ]

    operations = [
        migrations.AddField(
            model_name='servicesettings',
            name='content_type',
            field=models.ForeignKey(to='contenttypes.ContentType', null=True),
        ),
        migrations.AddField(
            model_name='servicesettings',
            name='object_id',
            field=models.PositiveIntegerField(null=True),
        ),
        migrations.AddField(
            model_name='servicesettings',
            name='tags',
            field=taggit.managers.TaggableManager(to='taggit.Tag', through='taggit.TaggedItem', blank=True, help_text='A comma-separated list of tags.', verbose_name='Tags'),
        ),
    ]



from __future__ import unicode_literals

from datetime import timedelta

from waldur_core.core import WaldurExtension


class MarketplaceExtension(WaldurExtension):
    class Settings:
        WALDUR_MARKETPLACE = {
            'THUMBNAIL_SIZE': (120, 120),
            'THUMBNAIL_SUFFIX': 'thumbnail',
            'OWNER_CAN_APPROVE_ORDER': True,
            'MANAGER_CAN_APPROVE_ORDER': False,
            'ADMIN_CAN_APPROVE_ORDER': False,
            'NOTIFY_STAFF_ABOUT_APPROVALS': False,
            'NOTIFY_ABOUT_RESOURCE_CHANGE': True,
            'DISABLE_SENDING_NOTIFICATIONS_ABOUT_RESOURCE_UPDATE': True,
            'OWNER_CAN_REGISTER_SERVICE_PROVIDER': False,
            'ORDER_LINK_TEMPLATE': 'https://www.example.com/#/projects/'
                                   '{project_uuid}/marketplace-order-list/',
            'ORDER_ITEM_LINK_TEMPLATE': 'https://www.example.com/#/projects/{project_uuid}/'
                                        'marketplace-order-item-details/{order_item_uuid}/'
        }

    @staticmethod
    def django_app():
        return 'waldur_mastermind.marketplace'

    @staticmethod
    def is_assembly():
        return True

    @staticmethod
    def get_public_settings():
        return [
            'OWNER_CAN_APPROVE_ORDER',
            'MANAGER_CAN_APPROVE_ORDER',
            'ADMIN_CAN_APPROVE_ORDER',
            'OWNER_CAN_REGISTER_SERVICE_PROVIDER',
        ]

    @staticmethod
    def django_urls():
        from .urls import urlpatterns
        return urlpatterns

    @staticmethod
    def rest_urls():
        from .urls import register_in
        return register_in

    @staticmethod
    def celery_tasks():
        return {
            'waldur-marketplace-calculate-usage': {
                'task': 'waldur_mastermind.marketplace.calculate_usage_for_current_month',
                'schedule': timedelta(hours=1),
                'args': (),
            }
        }



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.contrib.contenttypes.models import ContentType
from django.db import migrations


def copy_certifications_from_openstack_settings_to_openstack_tenant_settings(apps, schema_editor):
    ServiceSettings = apps.get_model('structure', 'ServiceSettings')
    Tenant = apps.get_model('openstack', 'Tenant')
    tenant_content_type = ContentType.objects.get_for_model(Tenant)
    openstack_tenant_settings = ServiceSettings.objects.filter(type='OpenStackTenant')

    for settings in openstack_tenant_settings.iterator():
        # skip all settings linked to different type than tenant content type.
        if settings.content_type_id != tenant_content_type.id:
            continue

        try:
            # GenericRelation is not available in migration, thus tenant has to be accessed directly through object_id
            tenant = Tenant.objects.get(pk=settings.object_id)
        except Tenant.DoesNotExist:
            continue
        else:
            admin_settings = tenant.service_project_link.service.settings
            settings.certifications.clear()
            settings.certifications.add(*admin_settings.certifications.all())


class Migration(migrations.Migration):

    dependencies = [
        ('openstack', '0001_initial'),
        ('openstack_tenant', '0024_add_backup_size'),
    ]

    operations = [
        migrations.RunPython(copy_certifications_from_openstack_settings_to_openstack_tenant_settings),
    ]



import operator

from waldur_core.core import models, utils, tasks


class BaseExecutor(object):
    """ Base class for describing logical operation with backend.

    Executor describes celery signature or primitive of low-level tasks that
    should be executed to provide high-level operation.

    Executor should handle:
     - low-level tasks execution;
     - models state changes;
    """

    @classmethod
    def get_task_signature(cls, instance, serialized_instance, **kwargs):
        """ Get Celery signature or chain that describes executor action.

        Each task should be subclass of LowLevelTask class.
        Celery Signature and Primitives:
         - http://docs.celeryproject.org/en/latest/userguide/canvas.html
        Examples:
         - to execute only one task - return Signature of necessary task: `task.si(serialized_instance)`
         - to execute several tasks - return Chain of tasks: `chain(t1.s(), t2.s())`
        Note! Celery chord and group is not supported.
        """
        raise NotImplementedError('Executor %s should implement method `get_task_signature`' % cls.__name__)

    @classmethod
    def get_success_signature(cls, instance, serialized_instance, **kwargs):
        """ Get Celery signature of task that should be applied on successful execution. """
        return None

    @classmethod
    def get_failure_signature(cls, instance, serialized_instance, **kwargs):
        """ Get Celery signature of task that should be applied on failed execution. """
        return None

    @classmethod
    def execute(cls, instance, async=True, countdown=2, is_heavy_task=False, **kwargs):
        """ Execute high level-operation """
        cls.pre_apply(instance, async=async, **kwargs)
        serialized_instance = utils.serialize_instance(instance)

        signature = cls.get_task_signature(instance, serialized_instance, **kwargs)
        link = cls.get_success_signature(instance, serialized_instance, **kwargs)
        link_error = cls.get_failure_signature(instance, serialized_instance, **kwargs)
        if async:
            return signature.apply_async(
                link=link,
                link_error=link_error,
                countdown=countdown,
                queue=is_heavy_task and 'heavy' or None
            )
        else:
            result = signature.apply()
            callback = link if not result.failed() else link_error
            if callback is not None:
                if not callback.immutable:
                    callback.args = (result.id,) + callback.args
                callback.apply()
            return result.get()  # wait until task is ready

    @classmethod
    def pre_apply(cls, instance, **kwargs):
        """ Perform synchronous actions before signature apply """
        pass

    @classmethod
    def as_signature(cls, instance, **kwargs):
        serialized_instance = utils.serialize_instance(instance)
        pre_apply = tasks.PreApplyExecutorTask().si(
            utils.serialize_class(cls), serialized_instance, **kwargs)
        main = cls.get_task_signature(instance, serialized_instance, **kwargs)
        link = cls.get_success_signature(instance, serialized_instance, **kwargs)
        link_error = cls.get_failure_signature(instance, serialized_instance, **kwargs)
        parts = [task for task in [pre_apply, main, link]
                 if not isinstance(task, tasks.EmptyTask) and task is not None]
        signature = reduce(operator.or_, parts)
        if link_error:
            signature = signature.on_error(link_error)
        return signature


class ExecutorException(Exception):
    pass


class ErrorExecutorMixin(object):
    """ Set object as erred on fail. """

    @classmethod
    def get_failure_signature(cls, instance, serialized_instance, **kwargs):
        return tasks.ErrorStateTransitionTask().s(serialized_instance)


class SuccessExecutorMixin(object):
    """ Set object as OK on success, cleanup action and its details. """

    @classmethod
    def get_success_signature(cls, instance, serialized_instance, **kwargs):
        return tasks.StateTransitionTask().si(
            serialized_instance, state_transition='set_ok', action='', action_details={})


class DeleteExecutorMixin(object):
    """ Delete object on success or if force flag is enabled """

    @classmethod
    def get_success_signature(cls, instance, serialized_instance, **kwargs):
        return tasks.DeletionTask().si(serialized_instance)

    @classmethod
    def get_failure_signature(cls, instance, serialized_instance, force=False, **kwargs):
        if force:
            return tasks.DeletionTask().si(serialized_instance)
        else:
            return tasks.ErrorStateTransitionTask().s(serialized_instance)


class EmptyExecutor(BaseExecutor):

    @classmethod
    def get_task_signature(cls, *args, **kwargs):
        return tasks.EmptyTask().si()


class CreateExecutor(SuccessExecutorMixin, ErrorExecutorMixin, BaseExecutor):
    """ Default states transition for object creation.

     - mark object as OK on success creation;
     - mark object as erred on failed creation;
    """
    pass


class UpdateExecutor(SuccessExecutorMixin, ErrorExecutorMixin, BaseExecutor):
    """ Default states transition for object update.

     - schedule updating before update;
     - mark object as OK on success update;
     - mark object as erred on failed update;
    """

    @classmethod
    def pre_apply(cls, instance, **kwargs):
        instance.schedule_updating()
        instance.save(update_fields=['state'])

    @classmethod
    def execute(cls, instance, async=True, **kwargs):
        if 'updated_fields' not in kwargs:
            raise ExecutorException('updated_fields keyword argument should be defined for UpdateExecutor.')
        super(UpdateExecutor, cls).execute(instance, async=async, **kwargs)


class DeleteExecutor(DeleteExecutorMixin, BaseExecutor):
    """ Default states transition for object deletion.

     - schedule deleting before deletion;
     - delete object on success deletion;
     - mark object as erred on failed deletion;
    """

    @classmethod
    def pre_apply(cls, instance, **kwargs):
        instance.schedule_deleting()
        instance.save(update_fields=['state'])


class ActionExecutor(SuccessExecutorMixin, ErrorExecutorMixin, BaseExecutor):
    """ Default states transition for executing action with object.

     - schedule updating before action execution;
     - mark object as OK on success action execution;
     - mark object as erred on failed action execution;
    """
    # TODO: After refactoring field action should become mandatory for implementation
    action = ''

    @classmethod
    def get_action_details(cls, instance, **kwargs):
        """ Get detailed action description """
        return {}

    @classmethod
    def pre_apply(cls, instance, **kwargs):
        if instance.state == models.StateMixin.States.UPDATE_SCHEDULED:
            return
        instance.schedule_updating()
        instance.action = cls.action
        instance.action_details = cls.get_action_details(instance, **kwargs)
        instance.save()



# -*- coding: utf-8 -*-
# Generated by Django 1.11.9 on 2018-03-27 14:01
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('python_management', '0003_added_unique_constraint'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='pythonmanagementdeleterequest',
            name='virtual_env_name',
        ),
        migrations.RemoveField(
            model_name='pythonmanagementfindvirtualenvsrequest',
            name='virtual_env_name',
        ),
    ]



from waldur_core.cost_tracking import CostTrackingRegister, CostTrackingStrategy, ConsumableItem

from . import models


class InstanceStrategy(CostTrackingStrategy):
    resource_class = models.Instance

    class Types(object):
        FLAVOR = 'flavor'

    @classmethod
    def get_consumable_items(cls):
        return [ConsumableItem(item_type=cls.Types.FLAVOR, key=size.backend_id, default_price=size.price)
                for size in models.Size.objects.all()]

    @classmethod
    def get_configuration(cls, instance):
        consumables = {}
        if instance.state != models.Instance.States.ERRED:
            consumables[ConsumableItem(item_type=cls.Types.FLAVOR, key=instance.size_backend_id)] = 1
        return consumables


CostTrackingRegister.register_strategy(InstanceStrategy)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations


def migrate_items(apps, schema_editor):
    OpenStackItem = apps.get_model('invoices', 'OpenStackItem')
    OfferingItem = apps.get_model('invoices', 'OfferingItem')
    GenericInvoiceItem = apps.get_model('invoices', 'GenericInvoiceItem')
    Offering = apps.get_model('support', 'Offering')
    OpenStackPackage = apps.get_model('packages', 'OpenStackPackage')

    ContentType = apps.get_model('contenttypes', 'ContentType')
    offering_ct = ContentType.objects.get_for_model(Offering)
    package_ct = ContentType.objects.get_for_model(OpenStackPackage)

    for item in OpenStackItem.objects.all():
        generic_item = GenericInvoiceItem.objects.create(
            invoice=item.invoice,
            details=item.package_details,
            start=item.start,
            end=item.end,
            project=item.project,
            project_name=item.project_name,
            project_uuid=item.project_uuid,
            unit_price=item.unit_price,
            unit=item.unit,
            product_code=item.product_code,
            article_code=item.article_code
        )
        if item.package:
            generic_item.content_type = package_ct
            generic_item.object_id = item.package.id
            generic_item.save()

    for item in OfferingItem.objects.all():
        generic_item = GenericInvoiceItem.objects.create(
            invoice=item.invoice,
            details=item.offering_details,
            start=item.start,
            end=item.end,
            project=item.project,
            project_name=item.project_name,
            project_uuid=item.project_uuid,
            unit_price=item.unit_price,
            unit=item.unit,
            product_code=item.product_code,
            article_code=item.article_code
        )
        if item.offering:
            generic_item.content_type = offering_ct
            generic_item.object_id = item.offering.id
            generic_item.save()


class Migration(migrations.Migration):

    dependencies = [
        ('invoices', '0026_invoice__file'),
        ('contenttypes', '0002_remove_content_type_name'),
    ]

    operations = [
        migrations.RunPython(migrate_items),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('openstack_tenant', '0008_backup_schedule'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='openstacktenantservice',
            options={'verbose_name': 'OpenStackTenant provider', 'verbose_name_plural': 'OpenStackTenant providers'},
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models
import waldur_core.core.fields
import waldur_core.core.validators


class Migration(migrations.Migration):

    dependencies = [
        ('structure', '0041_servicesettings_domain'),
        ('openstack_tenant', '0015_snapshotrestoration'),
    ]

    operations = [
        migrations.CreateModel(
            name='InternalIP',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('mac_address', models.CharField(max_length=32, blank=True)),
                ('ip4_address', models.GenericIPAddressField(null=True, protocol=b'IPv4', blank=True)),
                ('ip6_address', models.GenericIPAddressField(null=True, protocol=b'IPv6', blank=True)),
                ('instance', models.ForeignKey(related_name='internal_ips_set', to='openstack_tenant.Instance')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Network',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('description', models.CharField(max_length=500, verbose_name='description', blank=True)),
                ('name', models.CharField(max_length=150, verbose_name='name', validators=[waldur_core.core.validators.validate_name])),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('backend_id', models.CharField(max_length=255, db_index=True)),
                ('is_external', models.BooleanField(default=False)),
                ('type', models.CharField(max_length=50, blank=True)),
                ('segmentation_id', models.IntegerField(null=True)),
                ('settings', models.ForeignKey(related_name='+', to='structure.ServiceSettings')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='SubNet',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('description', models.CharField(max_length=500, verbose_name='description', blank=True)),
                ('name', models.CharField(max_length=150, verbose_name='name', validators=[waldur_core.core.validators.validate_name])),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('backend_id', models.CharField(max_length=255, db_index=True)),
                ('cidr', models.CharField(max_length=32, blank=True)),
                ('gateway_ip', models.GenericIPAddressField(null=True, protocol='IPv4')),
                ('allocation_pools', waldur_core.core.fields.JSONField(default={})),
                ('ip_version', models.SmallIntegerField(default=4)),
                ('enable_dhcp', models.BooleanField(default=True)),
                ('dns_nameservers', waldur_core.core.fields.JSONField(default=[], help_text='List of DNS name servers associated with the subnet.')),
                ('network', models.ForeignKey(related_name='subnets', to='openstack_tenant.Network')),
                ('settings', models.ForeignKey(related_name='+', to='structure.ServiceSettings')),
            ],
            options={
                'abstract': False,
                'verbose_name': 'Subnet',
                'verbose_name_plural': 'Subnets',
            },
        ),
        migrations.AddField(
            model_name='internalip',
            name='subnet',
            field=models.ForeignKey(related_name='internal_ips', to='openstack_tenant.SubNet'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.16 on 2018-11-23 09:33
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('marketplace', '0047_rewire_order_item'),
    ]

    operations = [
        migrations.AddField(
            model_name='cartitem',
            name='type',
            field=models.PositiveSmallIntegerField(choices=[(1, 'Create'), (2, 'Update'), (3, 'Terminate')], default=1),
        ),
        migrations.AddField(
            model_name='orderitem',
            name='type',
            field=models.PositiveSmallIntegerField(choices=[(1, 'Create'), (2, 'Update'), (3, 'Terminate')], default=1),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.13 on 2018-09-11 19:16
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('marketplace', '0032_orderitem_error_message'),
    ]

    operations = [
        migrations.AddField(
            model_name='plancomponent',
            name='billing_type',
            field=models.CharField(choices=[('fixed', 'Fixed-price'), ('usage', 'Usage-based')], default='fixed', max_length=5),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations


def expose_component_name(apps, schema_editor):
    ContentType = apps.get_model('contenttypes', 'ContentType')
    Offering = apps.get_model('support', 'Offering')
    GenericInvoiceItem = apps.get_model('invoices', 'GenericInvoiceItem')
    PlanComponent = apps.get_model('marketplace', 'PlanComponent')

    content_type = ContentType.objects.get_for_model(Offering)
    for item in GenericInvoiceItem.objects.filter(content_type=content_type):
        if item.details:
            plan_component_id = item.details.get('plan_component_id')
            if not plan_component_id:
                continue
            try:
                plan_component = PlanComponent.objects.get(id=plan_component_id)
            except PlanComponent.DoesNotExist:
                continue
            else:
                item.details['offering_component_name'] = plan_component.component.name
                item.details['plan_name'] = plan_component.plan.name
                item.save(update_fields=['details'])


class Migration(migrations.Migration):

    dependencies = [
        ('support_invoices', '0002_delete_requestbasedoffering'),
        ('contenttypes', '0002_remove_content_type_name'),
        ('support', '0036_offering_ordering'),
        ('invoices', '0030_json_details_on_genericinvoiceitem'),
        ('marketplace', '0078_fix_plan_component_amount'),
    ]

    operations = [
        migrations.RunPython(expose_component_name),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models
from decimal import Decimal
import django.core.validators

from .. import utils


class Migration(migrations.Migration):

    dependencies = [
        ('invoices', '0006_add_daily_price'),
    ]

    def migrate_daily_price(apps, schema_editor):
        OpenStackItem = apps.get_model('invoices', 'OpenStackItem')
        for item in OpenStackItem.objects.iterator():
            full_days = utils.get_full_days(item.start, item.end)
            daily_price = item.price / full_days if full_days else 0
            item.daily_price = daily_price
            item.save(update_fields=['daily_price'])

    operations = [
        migrations.RunPython(migrate_daily_price),
        migrations.RemoveField(
            model_name='openstackitem',
            name='price',
        ),
    ]



from __future__ import unicode_literals

import collections
from datetime import timedelta

from rest_framework import viewsets
from rest_framework.response import Response

from . import models, serializers


class DailyQuotaHistoryViewSet(viewsets.GenericViewSet):
    # Fix for schema generation
    queryset = []

    def list(self, request):
        serializer = serializers.DailyHistoryQuotaSerializer(
            data=request.query_params,
            context={'request': request},
        )
        serializer.is_valid(raise_exception=True)
        result = self.get_result(serializer.validated_data)
        return Response(result)

    def get_result(self, query):
        scope = query['scope']
        quota_names = query['quota_names']
        start = query['start']
        end = query['end']

        quotas = models.DailyQuotaHistory.objects.filter(
            scope=scope,
            name__in=quota_names,
            date__gte=start,
            date__lte=end,
        ).only(
            'name',
            'date',
            'usage',
        )
        charts = collections.defaultdict(dict)
        for quota in quotas:
            charts[quota.name][quota.date] = quota.usage

        values = collections.defaultdict(list)
        day = timedelta(days=1)
        days = (end - start).days
        for name in quota_names:
            usage = 0
            for i in range(days + 1):
                date = start + i * day
                usage = charts[name].get(date, usage)
                values[name].append(usage)
        return values



# -*- coding: utf-8 -*-
# Generated by Django 1.11.1 on 2017-06-07 08:52
from __future__ import unicode_literals

from django.db import migrations, models
import waldur_core.core.validators


class Migration(migrations.Migration):

    dependencies = [
        ('structure', '0051_add_customer_email_phone_agreement_number'),
    ]

    operations = [
        migrations.AddField(
            model_name='customer',
            name='access_subnets',
            field=models.TextField(blank=True, default='', validators=[waldur_core.core.validators.validate_cidr_list], help_text='Enter a comma separated list of IPv4 or IPv6 CIDR addresses from where connection to self-service is allowed.'),
        ),
    ]



import logging

from django.core import exceptions as django_exceptions
from django.db import transaction
from django.utils.timezone import now

from waldur_core.core.models import StateMixin

from . import log, models


logger = logging.getLogger(__name__)


def create_resource_plan_period(resource):
    models.ResourcePlanPeriod.objects.create(
        resource=resource,
        plan=resource.plan,
        start=now(),
        end=None,
    )


@transaction.atomic()
def close_resource_plan_period(resource):
    try:
        previous_period = models.ResourcePlanPeriod.objects.select_for_update().get(
            resource=resource,
            plan=resource.plan,
            end=None,
        )
        previous_period.end = now()
        previous_period.save(update_fields=['end'])
    except django_exceptions.ObjectDoesNotExist:
        logger.warning('Skipping previous resource plan period update '
                       'because it does not exist. Resource ID: %s, plan ID: %s.',
                       resource.id, resource.plan.id)
    except django_exceptions.MultipleObjectsReturned:
        logger.warning('Skipping previous resource plan period update '
                       'because multiple objects found. Resource ID: %s, plan ID: %s.',
                       resource.id, resource.plan.id)


def resource_creation_succeeded(resource):
    resource.set_state_ok()
    resource.save(update_fields=['state'])
    set_order_item_state(
        resource,
        models.RequestTypeMixin.Types.CREATE,
        models.OrderItem.States.DONE,
    )
    if resource.plan:
        create_resource_plan_period(resource)

    log.log_resource_creation_succeeded(resource)


def resource_creation_failed(resource):
    resource.set_state_erred()
    resource.save(update_fields=['state'])
    set_order_item_state(
        resource,
        models.RequestTypeMixin.Types.CREATE,
        models.OrderItem.States.ERRED,
    )

    log.log_resource_creation_failed(resource)


def resource_update_succeeded(resource):
    if resource.state != models.Resource.States.OK:
        resource.set_state_ok()
        resource.save(update_fields=['state'])
    order_item = set_order_item_state(
        resource,
        models.RequestTypeMixin.Types.UPDATE,
        models.OrderItem.States.DONE,
    )
    if order_item and order_item.plan:
        close_resource_plan_period(resource)

        resource.plan = order_item.plan
        resource.init_cost()
        resource.save(update_fields=['plan', 'cost'])

        create_resource_plan_period(resource)

    log.log_resource_update_succeeded(resource)


def resource_update_failed(resource):
    resource.set_state_erred()
    resource.save(update_fields=['state'])
    set_order_item_state(
        resource,
        models.RequestTypeMixin.Types.UPDATE,
        models.OrderItem.States.ERRED,
    )

    log.log_resource_update_failed(resource)


def resource_deletion_succeeded(resource):
    resource.set_state_terminated()
    resource.save(update_fields=['state'])
    set_order_item_state(
        resource,
        models.RequestTypeMixin.Types.TERMINATE,
        models.OrderItem.States.DONE,
    )

    if resource.plan:
        close_resource_plan_period(resource)

    log.log_resource_terminate_succeeded(resource)


def resource_deletion_failed(resource):
    resource.set_state_erred()
    resource.save(update_fields=['state'])
    set_order_item_state(
        resource,
        models.RequestTypeMixin.Types.TERMINATE,
        models.OrderItem.States.ERRED,
    )

    log.log_resource_terminate_failed(resource)


def set_order_item_state(resource, type, new_state):
    try:
        order_item = models.OrderItem.objects.get(
            resource=resource,
            type=type,
            state=models.OrderItem.States.EXECUTING,
        )
    except django_exceptions.ObjectDoesNotExist:
        logger.debug('Skipping order item synchronization for marketplace resource '
                     'because order item is not found. Resource ID: %s', resource.id)
    except django_exceptions.MultipleObjectsReturned:
        logger.debug('Skipping order item synchronization for marketplace resource '
                     'because there are multiple active order items are found. '
                     'Resource ID: %s', resource.id)
    else:
        getattr(order_item, OrderItemStateRouter[new_state])()
        order_item.save(update_fields=['state'])
        return order_item


States = StateMixin.States


StateRouter = {
    (States.CREATING, States.OK): resource_creation_succeeded,
    (States.CREATING, States.ERRED): resource_creation_failed,
    (States.UPDATING, States.OK): resource_update_succeeded,
    (States.UPDATING, States.ERRED): resource_update_failed,
    (States.DELETING, States.ERRED): resource_deletion_failed,
}


OrderItemStateRouter = {
    models.OrderItem.States.EXECUTING: 'set_state_executing',
    models.OrderItem.States.DONE: 'set_state_done',
    models.OrderItem.States.ERRED: 'set_state_erred',
    models.OrderItem.States.TERMINATED: 'set_state_terminated',
    models.OrderItem.States.TERMINATING: 'set_state_terminating',
}


def sync_resource_state(instance, resource):
    key = (instance.tracker.previous('state'), instance.state)
    func = StateRouter.get(key)
    if func:
        func(resource)



import django_filters

from . import models


class ProfileFilter(django_filters.FilterSet):
    user = django_filters.UUIDFilter(name='user__uuid')

    class Meta(object):
        model = models.Profile
        fields = []



from __future__ import unicode_literals

from django.db import transaction

from waldur_core.core import tasks as core_tasks, utils as core_utils
from waldur_openstack.openstack import models as openstack_models
from waldur_mastermind.packages import serializers

from .log import event_logger
from . import models


class OpenStackPackageErrorTask(core_tasks.ErrorStateTransitionTask):
    """ Handle OpenStack package creation errors.

        If error occurred on tenant creation - mark tenant and service settings as erred.
        If error occurred on service settings creation - mark only service settings as erred.
    """

    @classmethod
    def get_description(cls, result_id, package, *args, **kwargs):
        return 'Mark package "%s" components as erred.' % package

    def execute(self, package):
        if package.tenant.state != openstack_models.Tenant.States.OK:
            self.state_transition(package.tenant, 'set_erred')
            self.save_error_message(package.tenant)
            self.state_transition(package.service_settings, 'set_erred')
            package.service_settings.error_message = 'Failed to create tenant: %s.' % package.tenant
            package.service_settings.save(update_fields=['error_message'])
        else:
            self.state_transition(package.service_settings, 'set_erred')
            self.save_error_message(package.service_settings)


class OpenStackPackageSettingsPopulationTask(core_tasks.Task):
    """ Populate service settings options based on provisioned tenant. """

    @classmethod
    def get_description(cls, package, *args, **kwargs):
        return 'Copy tenant backend data to settings. Package "%s".' % package

    def execute(self, package):
        package.service_settings.options['tenant_id'] = package.tenant.backend_id
        package.service_settings.options['external_network_id'] = package.tenant.external_network_id
        package.service_settings.options['internal_network_id'] = package.tenant.internal_network_id
        package.service_settings.save()


class LogOpenStackPackageChange(core_tasks.Task):

    def execute(self, tenant, event, new_package, old_package, service_settings, *args, **kwargs):
        service_settings = core_utils.deserialize_instance(service_settings)
        event_type = 'openstack_package_change_succeeded' if event == 'succeeded' else 'openstack_package_change_failed'

        event_logger.openstack_package.info(
            'Tenant package changing has %s. '
            'Old value: %s, new value: {package_template_name}' % (event, old_package),
            event_type=event_type,
            event_context={
                'tenant': tenant,
                'package_template_name': new_package,
                'service_settings': service_settings,
            })


class OpenStackPackageSuccessTask(core_tasks.Task):
    def execute(self, tenant, serialized_new_template, serialized_old_package,
                serialized_service_settings, *args, **kwargs):
        new_template = core_utils.deserialize_instance(serialized_new_template)
        old_package = core_utils.deserialize_instance(serialized_old_package)
        service_settings = core_utils.deserialize_instance(serialized_service_settings)

        with transaction.atomic():
            serializers._set_tenant_quotas(tenant, new_template)
            serializers._set_related_service_settings_quotas(tenant, new_template)
            serializers._set_tenant_extra_configuration(tenant, new_template)
            old_package.delete()
            models.OpenStackPackage.objects.create(
                template=new_template,
                service_settings=service_settings,
                tenant=tenant
            )



from django.contrib import admin

from waldur_core.structure import admin as structure_admin

from .models import AWSService, AWSServiceProjectLink, Instance, Image, Region, Size, Volume


class ImageAdmin(structure_admin.BackendModelAdmin):
    fields = 'name', 'region', 'backend_id'
    list_display = 'name', 'region', 'backend_id'
    list_filter = 'region',


class ImageInline(admin.TabularInline):
    model = Image
    extra = 1


class RegionAdmin(structure_admin.ProtectedModelMixin, structure_admin.BackendModelAdmin):
    readonly_fields = 'name', 'backend_id'
    inlines = ImageInline,


class SizeAdmin(structure_admin.ProtectedModelMixin, structure_admin.BackendModelAdmin):
    readonly_fields = 'name', 'backend_id'
    list_display = 'name', 'backend_id', 'cores', 'ram', 'disk'


class VolumeAdmin(structure_admin.ResourceAdmin):
    pass


admin.site.register(Image, ImageAdmin)
admin.site.register(Size, SizeAdmin)
admin.site.register(Region, RegionAdmin)
admin.site.register(Instance, structure_admin.VirtualMachineAdmin)
admin.site.register(Volume, VolumeAdmin)
admin.site.register(AWSService, structure_admin.ServiceAdmin)
admin.site.register(AWSServiceProjectLink, structure_admin.ServiceProjectLinkAdmin)



from __future__ import unicode_literals

import collections
import json
import logging
from datetime import datetime

import dateutil.parser
from django.conf import settings
from django.utils import timezone
from django.core.exceptions import ObjectDoesNotExist
from django.db import transaction
from jira import Comment
from jira.utils import json_loads
from six.moves.html_parser import HTMLParser

from waldur_jira.backend import reraise_exceptions, JiraBackend
from waldur_mastermind.support import models
from waldur_mastermind.support.exceptions import SupportUserInactive

from . import SupportBackend

logger = logging.getLogger(__name__)


Settings = collections.namedtuple('Settings', ['backend_url', 'username', 'password', 'email', 'token'])


class ServiceDeskBackend(JiraBackend, SupportBackend):
    servicedeskapi_path = 'servicedeskapi'
    model_comment = models.Comment
    model_issue = models.Issue
    model_attachment = models.Attachment

    def __init__(self):
        self.settings = Settings(
            backend_url=settings.WALDUR_SUPPORT.get('CREDENTIALS', {}).get('server'),
            username=settings.WALDUR_SUPPORT.get('CREDENTIALS', {}).get('username'),
            password=settings.WALDUR_SUPPORT.get('CREDENTIALS', {}).get('password'),
            email=settings.WALDUR_SUPPORT.get('CREDENTIALS', {}).get('email'),
            token=settings.WALDUR_SUPPORT.get('CREDENTIALS', {}).get('token'),
        )
        self.verify = settings.WALDUR_SUPPORT.get('CREDENTIALS', {}).get('verify_ssl')
        self.project_settings = settings.WALDUR_SUPPORT.get('PROJECT', {})
        self.issue_settings = settings.WALDUR_SUPPORT.get('ISSUE', {})
        self.use_old_api = settings.WALDUR_SUPPORT.get('USE_OLD_API', False)

    def pull_service_properties(self):
        super(ServiceDeskBackend, self).pull_service_properties()
        self.pull_request_types()

    @reraise_exceptions
    def create_comment(self, comment):
        backend_comment = self._add_comment(
            comment.issue.backend_id,
            comment.prepare_message(),
            is_internal=not comment.is_public,
        )
        comment.backend_id = backend_comment.id
        comment.save(update_fields=['backend_id'])

    def _add_comment(self, issue, body, is_internal):
        data = {
            'body': body,
            'properties': [{'key': 'sd.public.comment', 'value': {'internal': is_internal}}, ]
        }

        url = self.manager._get_url('issue/{0}/comment'.format(issue))
        response = self.manager._session.post(url, data=json.dumps(data))

        comment = Comment(self.manager._options, self.manager._session, raw=json_loads(response))
        return comment

    @reraise_exceptions
    def create_issue(self, issue):
        if not issue.caller.email:
            return

        self.create_user(issue.caller)

        if self.use_old_api:
            return super(ServiceDeskBackend, self).create_issue(issue)

        args = self._issue_to_dict(issue)
        args['serviceDeskId'] = self.manager.service_desk(self.project_settings['key'])
        if not models.RequestType.objects.filter(issue_type_name=issue.type).count():
            self.pull_request_types()

        if not models.RequestType.objects.filter(issue_type_name=issue.type).count():
            logger.debug('Not exists a RequestType for this issue type %s', issue.type)
            return

        args['requestTypeId'] = models.RequestType.objects.filter(issue_type_name=issue.type).first().backend_id
        backend_issue = self.manager.create_customer_request(args)
        args = self._get_custom_fields(issue)

        # Update an issue, because create_customer_request doesn't allow setting custom fields.
        backend_issue.update(**args)
        self._backend_issue_to_issue(backend_issue, issue)
        issue.save()

    def create_user(self, user):
        # Temporary workaround as JIRA returns 500 error if user already exists
        if self.use_old_api:
            # old API has a bug that causes user active status to be set to False if includeInactive is passed as True
            existing_support_user = self.manager.search_users(user.email)
        else:
            existing_support_user = self.manager.search_users(user.email, includeInactive=True)

        if existing_support_user:
            active_user = [u for u in existing_support_user if u.active]
            if not active_user:
                raise SupportUserInactive()

            logger.debug('Skipping user %s creation because it already exists', user.email)
            backend_customer = active_user[0]
        else:
            if self.use_old_api:
                # add_user method returns boolean value therefore we need to fetch user object to find its key
                self.manager.add_user(user.email, user.email, fullname=user.full_name, ignore_existing=True)
                backend_customer = self.manager.search_users(user.email)[0]
            else:
                backend_customer = self.manager.create_customer(user.email, user.full_name)

        try:
            user.supportcustomer
        except ObjectDoesNotExist:
            support_customer = models.SupportCustomer(user=user, backend_id=backend_customer.key)
            support_customer.save()

    @reraise_exceptions
    def get_users(self):
        users = self.manager.search_assignable_users_for_projects('', self.project_settings['key'], maxResults=False)
        return [models.SupportUser(name=user.displayName, backend_id=user.key) for user in users]

    def _get_custom_fields(self, issue):
        args = {}

        if issue.reporter:
            args[self.get_field_id_by_name(self.issue_settings['reporter_field'])] = issue.reporter.name
        if issue.impact:
            args[self.get_field_id_by_name(self.issue_settings['impact_field'])] = issue.impact
        if issue.priority:
            args['priority'] = {'name': issue.priority}

        def set_custom_field(field_name, value):
            if value and self.issue_settings.get(field_name):
                args[self.get_field_id_by_name(self.issue_settings[field_name])] = value

        if issue.reporter and issue.reporter.user and issue.reporter.user.organization:
            set_custom_field('organisation_field', issue.reporter.user.organization)

        if issue.project:
            set_custom_field('project_field', issue.project.name)

        if issue.resource:
            set_custom_field('affected_resource_field', issue.resource)

        if issue.template:
            set_custom_field('template_field', issue.template.name)

        return args

    def _issue_to_dict(self, issue):
        parser = HTMLParser()

        if self.use_old_api:
            parser = HTMLParser()
            args = {
                'project': self.project_settings['key'],
                'summary': parser.unescape(issue.summary),
                'description': parser.unescape(issue.description),
                'issuetype': {'name': issue.type},
            }
            args.update(self._get_custom_fields(issue))

            try:
                support_user = models.SupportUser.objects.get(user=issue.caller)
                key = support_user.backend_id or issue.caller.email
            except models.SupportUser.DoesNotExist:
                key = issue.caller.email

            args[self.get_field_id_by_name(self.issue_settings['caller_field'])] = [{
                "name": key,  # will be equal to issue.caller.email for non-support users
                "key": key,
            }]
            return args

        args = {
            'requestFieldValues': {
                'summary': parser.unescape(issue.summary),
                'description': parser.unescape(issue.description)
            }
        }

        support_customer = issue.caller.supportcustomer
        args['requestParticipants'] = [support_customer.backend_id]
        return args

    def _get_first_sla_field(self, backend_issue):
        field_name = self.get_field_id_by_name(self.issue_settings['sla_field'])
        value = getattr(backend_issue.fields, field_name, None)
        if value and hasattr(value, 'ongoingCycle'):
            epoch_milliseconds = value.ongoingCycle.breachTime.epochMillis
            if epoch_milliseconds:
                return datetime.fromtimestamp(epoch_milliseconds / 1000.0, timezone.get_default_timezone())

    def _backend_issue_to_issue(self, backend_issue, issue):
        issue.key = backend_issue.key
        issue.backend_id = backend_issue.key
        issue.resolution = backend_issue.fields.resolution or ''
        issue.status = backend_issue.fields.status.name or ''
        issue.link = backend_issue.permalink()
        issue.priority = backend_issue.fields.priority.name
        issue.first_response_sla = self._get_first_sla_field(backend_issue)
        issue.summary = backend_issue.fields.summary
        issue.description = backend_issue.fields.description or ''
        issue.type = backend_issue.fields.issuetype.name
        issue.resolution_date = backend_issue.fields.resolutiondate or None

        def get_support_user_by_field(fields, field_name):
            support_user = None
            backend_user = getattr(fields, field_name, None)

            if backend_user:
                try:
                    support_user_backend_key = getattr(backend_user, 'key', None)
                    if support_user_backend_key:
                        support_user, _ = models.SupportUser.objects.get_or_create(backend_id=support_user_backend_key)

                except TypeError:
                    # except TypeError because 'item in self.raw' in here jira/resources.py:173
                    pass

            return support_user

        impact_field_id = self.get_field_id_by_name(self.issue_settings['impact_field'])
        impact = getattr(backend_issue.fields, impact_field_id, None)
        if impact:
            issue.impact = impact

        assignee = get_support_user_by_field(backend_issue.fields, 'assignee')
        if assignee:
            issue.assignee = assignee

        reporter = get_support_user_by_field(backend_issue.fields, 'reporter')
        if reporter:
            issue.reporter = reporter

    def _get_author(self, resource):
        backend_id = resource.raw.get('author', {}).get('key')
        author, _ = models.SupportUser.objects.get_or_create(backend_id=backend_id)
        return author

    def _backend_comment_to_comment(self, backend_comment, comment):
        comment.update_message(backend_comment.body)
        author = self._get_author(backend_comment)
        comment.author = author
        internal = self._get_property('comment', backend_comment.id, 'sd.public.comment')
        comment.is_public = not internal.get('value', {}).get('internal', False)

    def _backend_attachment_to_attachment(self, backend_attachment, attachment):
        author = self._get_author(backend_attachment)
        attachment.mime_type = getattr(backend_attachment, 'mimeType', '')
        attachment.file_size = backend_attachment.size
        attachment.created = dateutil.parser.parse(backend_attachment.created)
        attachment.author = author

    @reraise_exceptions
    def pull_request_types(self):
        service_desk_id = self.manager.service_desk(self.project_settings['key'])
        backend_request_types = self.manager.request_types(service_desk_id)
        with transaction.atomic():
            backend_request_type_map = {
                int(request_type.id): request_type for request_type in backend_request_types
            }

            waldur_request_type = {
                request_type.backend_id: request_type
                for request_type in models.RequestType.objects.all()
            }

            stale_request_types = set(waldur_request_type.keys()) - set(backend_request_type_map.keys())
            models.RequestType.objects.filter(backend_id__in=stale_request_types).delete()

            for backend_request_type in backend_request_types:
                issue_type = self.manager.issue_type(backend_request_type.issueTypeId)
                models.RequestType.objects.update_or_create(
                    backend_id=backend_request_type.id,
                    defaults={
                        'name': backend_request_type.name,
                        'issue_type_name': issue_type.name
                    })



from django.db.models import Sum

from waldur_slurm import models, utils


def get_usage():
    if not models.Allocation.objects.exists():
        return []
    qs = models.Allocation.objects.all()
    params = dict(('total_%s' % quota, Sum(quota)) for quota in utils.FIELD_NAMES)
    qs = qs.aggregate(**params)

    points = []
    for quota in utils.FIELD_NAMES:
        points.append({
            'measurement': 'slurm_%s' % quota,
            'fields': {
                'value': qs['total_%s' % quota]
            }
        })

    return points



from django.utils import timezone

from waldur_core.structure.permissions import _get_project
from waldur_mastermind.invoices import models as invoices_models, utils as invoices_utils
from waldur_mastermind.invoices.registrators import BaseRegistrator
from waldur_mastermind.marketplace import utils as marketplace_utils
from waldur_mastermind.packages import models as packages_models

from . import utils


class OpenStackItemRegistrator(BaseRegistrator):

    def get_customer(self, source):
        return source.tenant.service_project_link.project.customer

    def get_sources(self, customer):
        return packages_models.OpenStackPackage.objects.filter(
            tenant__service_project_link__project__customer=customer).distinct()

    def _find_item(self, source, now):
        result = utils.get_openstack_items().filter(
            object_id=source.id,
            invoice__customer=self.get_customer(source),
            invoice__state=invoices_models.Invoice.States.PENDING,
            invoice__year=now.year,
            invoice__month=now.month,
        ).first()
        return result

    def _create_item(self, source, invoice, start, end):
        package = source
        overlapping_item = utils.get_openstack_items().filter(
            invoice=invoice,
            end__day=start.day,
            details__tenant_name=package.tenant.name,
        ).order_by('-unit_price').first()

        daily_price = package.template.price
        product_code = package.template.product_code
        article_code = package.template.article_code
        if overlapping_item:
            """
            Notes:
            |- date -| - used during the date
            |- **** -| - used during the day
            |- ---- -| - was requested to use in the current day but will be moved to next or previous one.
            |-***?---| - was used for a half day and '?' stands for a conflict.

            If there is an item that overlaps with current one as shown below:
            |--03.01.2017-|-********-|-***?---|
                                     |----?**-|-06.01.2017-|-******-|
            we have to make next steps:
            1) If item is more expensive -> use it for price calculation
                and register new package starting from next day [-06.01.2017-]
            |--03.01.2017-|-********-|-*****-|
                                     |-------|-06.01.2017-|-******-|

            2) If old package item is more expensive and it is the end of the month
            extend package usage till the end of the day and set current package end date to start date,
            so that usage days is 0 but it is still registered in the invoice.
            |--29.01.2017-|-********-|-***31.01.2017***-|
                                     |----31.01.2017----|

            3) If item is cheaper do exactly the opposite and shift its end date to yesterday,
            so new package will be registered today
            |--03.01.2017-|-********-|-------|
                                     |-*****-|-06.01.2017-|-******-|
            """
            if overlapping_item.unit_price > daily_price:
                if overlapping_item.end.day == invoices_utils.get_current_month_end().day:
                    utils.extend_to_the_end_of_the_day(overlapping_item)
                    end = start
                else:
                    start = start + timezone.timedelta(days=1)
            else:
                utils.shift_backward(overlapping_item)

        invoices_models.GenericInvoiceItem.objects.create(
            scope=package,
            project=_get_project(package),
            unit_price=daily_price,
            unit=invoices_models.GenericInvoiceItem.Units.PER_DAY,
            product_code=product_code,
            article_code=article_code,
            invoice=invoice,
            start=start,
            end=end,
            details=self.get_details(package))

    def get_details(self, source):
        package = source
        details = {
            'name': utils.get_invoice_item_name(package),
            'tenant_name': package.tenant.name,
            'tenant_uuid': package.tenant.uuid.hex,
            'template_name': package.template.name,
            'template_uuid': package.template.uuid.hex,
            'template_category': package.template.get_category_display(),
        }
        service_provider_info = marketplace_utils.get_service_provider_info(source)
        details.update(service_provider_info)
        return details

    def get_name(self, source):
        return utils.get_invoice_item_name(source)

    def terminate(self, source, now=None):
        super(OpenStackItemRegistrator, self).terminate(source, now)
        package = source
        utils.get_openstack_items().filter(object_id=package.id).update(object_id=None)



from __future__ import unicode_literals

from waldur_core.core import WaldurExtension


class MarketplaceSupportExtension(WaldurExtension):
    class Settings:
        WALDUR_MARKETPLACE_SUPPORT = {
            'REQUEST_LINK_TEMPLATE': 'https://www.example.com/#/offering/{request_uuid}/',
            'CREATE_RESOURCE_TEMPLATE': (
                '\n[Order item|{{order_item_url}}].'
                '\nVendor: {{order_item.offering.customer.name}}'
            ),
            'UPDATE_RESOURCE_TEMPLATE': (
                '\n[Switch plan for resource {{order_item.resource.scope.name}}|{{request_url}}].'
                '\nSwitch from {{order_item.resource.plan.name}} plan to {{order_item.plan.name}}.'
            ),
            'TERMINATE_RESOURCE_TEMPLATE': (
                '[Terminate resource {{order_item.resource.scope.name}}|{{request_url}}].'
            ),
        }

    @staticmethod
    def django_app():
        return 'waldur_mastermind.marketplace_support'

    @staticmethod
    def is_assembly():
        return True



from collections import OrderedDict
from operator import itemgetter

from django.urls import NoReverseMatch
from rest_framework import views
from rest_framework.response import Response
from rest_framework.reverse import reverse
from rest_framework.routers import DefaultRouter


class SortedDefaultRouter(DefaultRouter):

    def get_api_root_view(self, api_urls=None):
        """
        Return a basic root view.
        """
        api_root_dict = OrderedDict()
        list_name = self.routes[0].name
        for prefix, viewset, basename in self.registry:
            api_root_dict[prefix] = list_name.format(basename=basename)

        class APIRootView(views.APIView):
            _ignore_model_permissions = True
            exclude_from_schema = True

            def get(self, request, *args, **kwargs):
                # Return a plain {"name": "hyperlink"} response.
                ret = OrderedDict()
                namespace = request.resolver_match.namespace
                for key, url_name in sorted(api_root_dict.items(), key=itemgetter(0)):
                    if namespace:
                        url_name = namespace + ':' + url_name
                    try:
                        ret[key] = reverse(
                            url_name,
                            args=args,
                            kwargs=kwargs,
                            request=request,
                            format=kwargs.get('format', None)
                        )
                    except NoReverseMatch:
                        # Don't bail out if eg. no list routes exist, only detail routes.
                        continue

                return Response(ret)

        return APIRootView.as_view()

    def get_default_base_name(self, viewset):
        """
        Attempt to automatically determine base name using `get_url_name`.
        """
        queryset = getattr(viewset, 'queryset', None)

        if queryset is not None:
            get_url_name = getattr(queryset.model, 'get_url_name', None)
            if get_url_name is not None:
                return get_url_name()

        return super(SortedDefaultRouter, self).get_default_base_name(viewset)

    def get_method_map(self, viewset, method_map):
        # head method is not included by default
        mappings = super(SortedDefaultRouter, self).get_method_map(viewset, method_map)
        if 'get' in mappings:
            mappings['head'] = mappings['get']

        return mappings



from __future__ import unicode_literals

from collections import defaultdict
from functools import reduce
import inspect
import logging

from django.contrib.contenttypes import fields as ct_fields
from django.contrib.contenttypes import models as ct_models
from django.db import models, transaction
from django.db.models import Sum
from django.utils.encoding import python_2_unicode_compatible
from django.utils.translation import ugettext_lazy as _
from model_utils import FieldTracker
from reversion import revisions as reversion
import six

from waldur_core.core.models import UuidMixin, ReversionMixin, DescendantMixin
from waldur_core.logging.loggers import LoggableMixin
from waldur_core.logging.models import AlertThresholdMixin
from waldur_core.quotas import exceptions, managers, fields

logger = logging.getLogger(__name__)


@python_2_unicode_compatible
@reversion.register(fields=['usage', 'limit'])
class Quota(UuidMixin, AlertThresholdMixin, LoggableMixin, ReversionMixin, models.Model):
    """
    Abstract quota for any resource.

    Quota can exist without scope: for example, a quota for all projects or all
    customers on site.
    If quota limit is set to -1 quota will never be exceeded.
    """
    class Meta:
        unique_together = (('name', 'content_type', 'object_id'),)

    limit = models.FloatField(default=-1)
    usage = models.FloatField(default=0)
    name = models.CharField(max_length=150, db_index=True)

    content_type = models.ForeignKey(ct_models.ContentType, null=True)
    object_id = models.PositiveIntegerField(null=True)
    scope = ct_fields.GenericForeignKey('content_type', 'object_id')

    objects = managers.QuotaManager('scope')
    tracker = FieldTracker()

    def __str__(self):
        return '%s quota for %s' % (self.name, self.scope)

    def is_exceeded(self, delta=None, threshold=None):
        """
        Check is quota exceeded

        If delta is not None then checks if quota exceeds with additional delta usage
        If threshold is not None then checks if quota usage over threshold * limit
        """
        if self.limit == -1:
            return False

        # Allow to decrease quota usage
        if delta < 0:
            return False

        usage = self.usage
        limit = self.limit

        if delta is not None:
            usage += delta
        if threshold is not None:
            limit = threshold * limit

        return usage > limit

    def get_log_fields(self):
        return ('uuid', 'name', 'limit', 'usage', 'scope')

    def get_field(self):
        fields = self.scope.get_quotas_fields()
        try:
            return next(f for f in fields if f.name == self.name)
        except StopIteration:
            return

    def is_over_threshold(self):
        return self.usage >= self.threshold


class QuotaModelMixin(models.Model):
    """
    Add general fields and methods to model for quotas usage.

    Model with quotas have inherit this mixin.
    For quotas implementation such methods and fields have to be defined:
      - class Quota(QuotaModelMixin) - class with quotas fields as attributes.
      - can_user_update_quotas(self, user) - Return True if user has permission to update quotas of this object.
      - GLOBAL_COUNT_QUOTA_NAME - Name of global count quota. It presents - global quota will be automatically created
                                  for model. Optional attribute.

    Example:
        Customer(models.Model):
            ...
            Quotas(quotas_models.QuotaModelMixin.Quotas):
                nc_user_count = quotas_fields.QuotaField()  # define user count quota for customers

            # optional descriptor to direct access to quota
            nc_user_count = quotas_fields.QuotaLimitField(quota_field=Quotas.nc_user_count)

            def can_user_update_quotas(self, user):
                # only staff user can edit Customer quotas
                return user.is_staff

    Use such methods to change objects quotas:
      set_quota_limit, set_quota_usage, add_quota_usage.

    Helper methods validate_quota_change and get_sum_of_quotas_as_dict provide common operations with objects quotas.
    Check methods docstrings for more details.
    """

    class Quotas(six.with_metaclass(fields.FieldsContainerMeta)):
        enable_fields_caching = True
        # register model quota fields here

    class Meta:
        abstract = True

    quotas = ct_fields.GenericRelation('quotas.Quota', related_query_name='quotas')

    def get_or_create_quota(self, quota_name_or_field):
        if isinstance(quota_name_or_field, six.string_types):
            quota_name_or_field = getattr(self.Quotas, quota_name_or_field)
        quota, _ = quota_name_or_field.get_or_create_quota(self)
        return quota

    @transaction.atomic
    def set_quota_limit(self, quota_name, limit):
        quota = self.get_or_create_quota(quota_name)
        if quota.limit != limit:
            quota.limit = limit
            quota.save(update_fields=['limit'])

    @transaction.atomic
    def set_quota_usage(self, quota_name, usage):
        quota = self.get_or_create_quota(quota_name)
        if quota.usage != usage:
            quota.usage = usage
            quota.save(update_fields=['usage'])

    @transaction.atomic
    def add_quota_usage(self, quota_name, usage_delta, validate=False):
        quota = self.get_or_create_quota(quota_name)
        if validate and quota.is_exceeded(usage_delta):
            raise exceptions.QuotaValidationError(
                _('%(quota)s "%(name)s" quota is over limit. Required: %(usage)s, limit: %(limit)s.') % dict(
                    quota=self, name=quota_name, usage=quota.usage + usage_delta, limit=quota.limit))
        quota.usage += usage_delta
        if quota.usage < 0:
            logger.error('%(quota)s "%(name)s" quota usage should not be negative. '
                         'Current usage: %(usage)s, delta: %(usage_delta)s',
                         dict(quota=self, name=quota_name, usage=quota.usage, usage_delta=usage_delta))
            quota.usage = 0
        quota.save(update_fields=['usage'])

    def get_quota_ancestors(self):
        if isinstance(self, DescendantMixin):
            # We need to use set in order to eliminate duplicates.
            # Consider, for example, two ways of traversing from resource to customer:
            # resource -> spl -> project -> customer
            # resource -> spl -> service -> customer
            return {a for a in self.get_ancestors() if isinstance(a, QuotaModelMixin)}
        return {}

    def validate_quota_change(self, quota_deltas, raise_exception=False):
        """
        Get error messages about object and his ancestor quotas that will be exceeded if quota_delta will be added.

        raise_exception - if True QuotaExceededException will be raised if validation fails
        quota_deltas - dictionary of quotas deltas, example:
        {
            'ram': 1024,
            'storage': 2048,
            ...
        }
        Example of output:
            ['ram quota limit: 1024, requires: 2048(instance#1)', ...]

        """
        errors = []
        for name, delta in six.iteritems(quota_deltas):
            quota = self.quotas.get(name=name)
            if quota.is_exceeded(delta):
                errors.append('%s quota limit: %s, requires %s (%s)\n' % (
                    quota.name, quota.limit, quota.usage + delta, quota.scope))
        if not raise_exception:
            return errors
        else:
            if errors:
                raise exceptions.QuotaExceededException(_('One or more quotas were exceeded: %s') % ';'.join(errors))

    def can_user_update_quotas(self, user):
        """
        Return True if user has permission to update quota
        """
        return user.is_staff

    @classmethod
    def get_sum_of_quotas_as_dict(cls, scopes, quota_names=None, fields=['usage', 'limit']):
        """
        Return dictionary with sum of all scopes' quotas.

        Dictionary format:
        {
            'quota_name1': 'sum of limits for quotas with such quota_name1',
            'quota_name1_usage': 'sum of usages for quotas with such quota_name1',
            ...
        }
        All `scopes` have to be instances of the same model.
        `fields` keyword argument defines sum of which fields of quotas will present in result.
        """
        if not scopes:
            return {}

        if quota_names is None:
            quota_names = cls.get_quotas_names()

        scope_models = set([scope._meta.model for scope in scopes])
        if len(scope_models) > 1:
            raise exceptions.QuotaError(_('All scopes have to be instances of the same model.'))

        filter_kwargs = {
            'content_type': ct_models.ContentType.objects.get_for_model(scopes[0]),
            'object_id__in': [scope.id for scope in scopes],
            'name__in': quota_names
        }

        result = {}
        if 'usage' in fields:
            items = Quota.objects.filter(**filter_kwargs)\
                         .values('name').annotate(usage=Sum('usage'))
            for item in items:
                result[item['name'] + '_usage'] = item['usage']

        if 'limit' in fields:
            unlimited_quotas = Quota.objects.filter(limit=-1, **filter_kwargs)
            unlimited_quotas = list(unlimited_quotas.values_list('name', flat=True))
            for quota_name in unlimited_quotas:
                result[quota_name] = -1

            items = Quota.objects\
                         .filter(**filter_kwargs)\
                         .exclude(name__in=unlimited_quotas)\
                         .values('name')\
                         .annotate(limit=Sum('limit'))
            for item in items:
                result[item['name']] = item['limit']

        return result

    @classmethod
    def get_sum_of_quotas_for_querysets(cls, querysets, quota_names=None):
        partial_sums = [qs.model.get_sum_of_quotas_as_dict(qs, quota_names) for qs in querysets]
        return reduce(cls._sum_dicts, partial_sums, defaultdict(lambda: 0))

    @classmethod
    def _sum_dicts(cls, total, partial):
        for key, val in partial.items():
            if val != -1:
                total[key] += val
        for key in partial.keys():
            if key not in total:
                total[key] = -1
        return total

    @classmethod
    def get_quotas_fields(cls, field_class=None):
        if not hasattr(cls, '_quota_fields') or not cls.Quotas.enable_fields_caching:
            cls._quota_fields = dict(inspect.getmembers(cls.Quotas, lambda m: isinstance(m, fields.QuotaField))).values()
        if field_class is not None:
            return [v for v in cls._quota_fields if isinstance(v, field_class)]
        return cls._quota_fields

    @classmethod
    def get_quotas_names(cls):
        return [f.name for f in cls.get_quotas_fields()]


class ExtendableQuotaModelMixin(QuotaModelMixin):
    """ Allows to add quotas to model in runtime.

    Example:
        from waldur_core.quotas.fields import QuotaField

        QuotaScopeModel.add_quota_field(
            name='quota_name',
            quota_field=QuotaField(...),
        )
    """

    class Quotas(QuotaModelMixin.Quotas):
        enable_fields_caching = False
        # register model quota fields here

    class Meta:
        abstract = True

    @classmethod
    def add_quota_field(cls, name, quota_field):
        # We need to initiate name field here because quota is not listed in Quotas class
        # and initialization is not executed automatically.
        quota_field.name = name
        setattr(cls.Quotas, name, quota_field)
        from waldur_core.quotas.apps import QuotasConfig
        # For counter quotas we need to register signals explicitly
        if isinstance(quota_field, fields.CounterQuotaField):
            QuotasConfig.register_counter_field_signals(model=cls, counter_field=quota_field)


class SharedQuotaMixin(object):
    """
    This mixin updates quotas for several scopes.
    """

    def get_quota_deltas(self):
        """
        This method should return dict where key is quota name and value is quota diff.
        For example:
        {
            'storage': 1024,
            'volumes': 1,
        }
        """
        raise NotImplementedError()

    def get_quota_scopes(self):
        """
        This method should return list of quota model mixins.
        """
        raise NotImplementedError()

    def apply_quota_changes(self, validate=False, mult=1):
        scopes = self.get_quota_scopes()
        deltas = self.get_quota_deltas()
        for name, delta in deltas.items():
            for scope in scopes:
                if scope:
                    scope.add_quota_usage(name, delta * mult, validate=validate)

    def increase_backend_quotas_usage(self, validate=True):
        self.apply_quota_changes(validate=validate)

    def decrease_backend_quotas_usage(self):
        self.apply_quota_changes(mult=-1)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.18 on 2019-02-12 12:12
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('analytics', '0002_import_quotas'),
    ]

    operations = [
        migrations.AlterField(
            model_name='dailyquotahistory',
            name='content_type',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='contenttypes.ContentType'),
        ),
        migrations.AlterField(
            model_name='dailyquotahistory',
            name='object_id',
            field=models.PositiveIntegerField(),
        ),
        migrations.AlterUniqueTogether(
            name='dailyquotahistory',
            unique_together=set([('content_type', 'object_id', 'name', 'date')]),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.18 on 2019-02-25 09:10
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('marketplace', '0056_plan_max_amount'),
    ]

    operations = [
        migrations.AddField(
            model_name='category',
            name='backend_id',
            field=models.CharField(blank=True, max_length=255),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.16 on 2018-12-28 12:51
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('marketplace', '0049_offering_billable'),
    ]

    operations = [
        migrations.AddField(
            model_name='offeringcomponent',
            name='limit_amount',
            field=models.IntegerField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='offeringcomponent',
            name='limit_period',
            field=models.CharField(blank=True, choices=[('month', 'Maximum monthly - every month service provider can report up to the amount requested by user.'), ('total', 'Maximum total - SP can report up to the requested amount over the whole active state of resource.')], max_length=5, null=True),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.18 on 2019-03-12 11:26
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('marketplace', '0063_add_scope_for_resource_count'),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name='aggregateresourcecount',
            unique_together=set([('category', 'content_type', 'object_id')]),
        ),
        migrations.RemoveField(
            model_name='aggregateresourcecount',
            name='project',
        ),
    ]



from __future__ import unicode_literals

from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ObjectDoesNotExist
from django.db.models import Prefetch
from rest_framework import viewsets, exceptions

from waldur_core.core import views as core_views
from waldur_core.cost_tracking import models, serializers, filters
from waldur_core.structure import SupportedServices
from waldur_core.structure import models as structure_models, permissions as structure_permissions
from waldur_core.structure.filters import ScopeTypeFilterBackend


class PriceEstimateViewSet(core_views.ReadOnlyActionsViewSet):
    queryset = models.PriceEstimate.objects.all()
    serializer_class = serializers.PriceEstimateSerializer
    lookup_field = 'uuid'
    filter_backends = (
        filters.PriceEstimateDateFilterBackend,
        filters.PriceEstimateCustomerFilterBackend,
        filters.PriceEstimateScopeFilterBackend,
        ScopeTypeFilterBackend,
    )

    def get_serializer_context(self):
        context = super(PriceEstimateViewSet, self).get_serializer_context()
        try:
            depth = int(self.request.query_params['depth'])
        except (TypeError, KeyError):
            pass  # use default depth if it is not defined or defined wrongly.
        else:
            context['depth'] = min(depth, 10)  # DRF restriction - serializer depth cannot be > 10
        return context

    def get_queryset(self):
        return models.PriceEstimate.objects.filtered_for_user(self.request.user).order_by(
            '-year', '-month')

    def list(self, request, *args, **kwargs):
        """
        To get a list of price estimates, run **GET** against */api/price-estimates/* as authenticated user.
        You can filter price estimates by scope type, scope URL, customer UUID.

        `scope_type` is generic type of object for which price estimate is calculated.
        Currently there are following types: customer, project, service, serviceprojectlink, resource.

        `date` parameter accepts list of dates. `start` and `end` parameters together specify date range.
        Each valid date should in format YYYY.MM

        You can specify GET parameter ?depth to show price estimate children. For example with ?depth=2 customer
        price estimate will shows its children - project and service and grandchildren - serviceprojectlink.
        """
        return super(PriceEstimateViewSet, self).list(request, *args, **kwargs)


class PriceListItemViewSet(viewsets.ModelViewSet):
    queryset = models.PriceListItem.objects.all()
    serializer_class = serializers.PriceListItemSerializer
    lookup_field = 'uuid'
    filter_backends = (filters.PriceListItemServiceFilterBackend,)

    def get_queryset(self):
        return models.PriceListItem.objects.filtered_for_user(self.request.user)

    def _user_can_modify_price_list_item(self, item):
        if self.request.user.is_staff:
            return True

        customer = structure_permissions._get_customer(item)
        return customer.has_user(self.request.user, structure_models.CustomerRole.OWNER)

    def list(self, request, *args, **kwargs):
        """
        To get a list of price list items, run **GET** against */api/price-list-items/* as an authenticated user.
        """
        return super(PriceListItemViewSet, self).list(request, *args, **kwargs)

    def create(self, request, *args, **kwargs):
        """
        Run **POST** request against */api/price-list-items/* to create new price list item.
        Customer owner and staff can create price items.

        Example of request:

        .. code-block:: http

            POST /api/price-list-items/ HTTP/1.1
            Content-Type: application/json
            Accept: application/json
            Authorization: Token c84d653b9ec92c6cbac41c706593e66f567a7fa4
            Host: example.com

            {
                "units": "per month",
                "value": 100,
                "service": "http://example.com/api/oracle/d4060812ca5d4de390e0d7a5062d99f6/",
                "default_price_list_item": "http://example.com/api/default-price-list-items/349d11e28f634f48866089e41c6f71f1/"
            }
        """
        return super(PriceListItemViewSet, self).create(request, *args, **kwargs)

    def update(self, request, *args, **kwargs):
        """
        Run **PATCH** request against */api/price-list-items/<uuid>/* to update price list item.
        Only item_type, key value and units can be updated.
        Only customer owner and staff can update price items.
        """
        return super(PriceListItemViewSet, self).update(request, *args, **kwargs)

    def destroy(self, request, *args, **kwargs):
        """
        Run **DELETE** request against */api/price-list-items/<uuid>/* to delete price list item.
        Only customer owner and staff can delete price items.
        """
        return super(PriceListItemViewSet, self).destroy(request, *args, **kwargs)

    def initial(self, request, *args, **kwargs):
        if self.action in ('partial_update', 'update', 'destroy'):
            price_list_item = self.get_object()
            if not self._user_can_modify_price_list_item(price_list_item.service):
                raise exceptions.PermissionDenied()

        return super(PriceListItemViewSet, self).initial(request, *args, **kwargs)

    def perform_create(self, serializer):
        if not self._user_can_modify_price_list_item(serializer.validated_data['service']):
            raise exceptions.PermissionDenied()

        super(PriceListItemViewSet, self).perform_create(serializer)


class DefaultPriceListItemViewSet(viewsets.ReadOnlyModelViewSet):
    queryset = models.DefaultPriceListItem.objects.all()
    lookup_field = 'uuid'
    filter_class = filters.DefaultPriceListItemFilter
    serializer_class = serializers.DefaultPriceListItemSerializer

    def list(self, request, *args, **kwargs):
        """
        To get a list of default price list items, run **GET** against */api/default-price-list-items/*
        as authenticated user.

        Price lists can be filtered by:
         - ?key=<string>
         - ?item_type=<string> has to be from list of available item_types
           (available options: 'flavor', 'storage', 'license-os', 'license-application', 'network', 'support')
         - ?resource_type=<string> resource type, for example: 'OpenStack.Instance, 'Oracle.Database')
        """
        return super(DefaultPriceListItemViewSet, self).list(request, *args, **kwargs)


class MergedPriceListItemViewSet(viewsets.ReadOnlyModelViewSet):
    lookup_field = 'uuid'
    filter_class = filters.DefaultPriceListItemFilter
    serializer_class = serializers.MergedPriceListItemSerializer

    def list(self, request, *args, **kwargs):
        """
        To get a list of price list items, run **GET** against */api/merged-price-list-items/*
        as authenticated user.

        If service is not specified default price list items are displayed.
        Otherwise service specific price list items are displayed.
        In this case rendered object contains {"is_manually_input": true}

        In order to specify service pass query parameters:
        - service_type (Azure, OpenStack etc.)
        - service_uuid

        Example URL: http://example.com/api/merged-price-list-items/?service_type=Azure&service_uuid=cb658b491f3644a092dd223e894319be
        """
        return super(MergedPriceListItemViewSet, self).list(request, *args, **kwargs)

    def get_queryset(self):
        queryset = models.DefaultPriceListItem.objects.all()
        service = self._find_service()
        if service:
            # Filter items by resource type
            resources = SupportedServices.get_related_models(service)['resources']
            content_types = ContentType.objects.get_for_models(*resources).values()
            queryset = queryset.filter(resource_content_type__in=content_types)

            # Attach service-specific items
            price_list_items = models.PriceListItem.objects.filter(service=service)
            prefetch = Prefetch('pricelistitem_set', queryset=price_list_items, to_attr='service_item')
            queryset = queryset.prefetch_related(prefetch)
        return queryset

    def _find_service(self):
        service_type = self.request.query_params.get('service_type')
        service_uuid = self.request.query_params.get('service_uuid')
        if not service_type or not service_uuid:
            return
        rows = SupportedServices.get_service_models()
        if service_type not in rows:
            return
        service_class = rows.get(service_type)['service']
        try:
            return service_class.objects.get(uuid=service_uuid)
        except ObjectDoesNotExist:
            return None



# -*- coding: utf-8 -*-
# Generated by Django 1.11.7 on 2018-04-26 09:23
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('waldur_jira', '0015_attachment_thumbnail'),
    ]

    operations = [
        migrations.AlterField(
            model_name='project',
            name='template',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='waldur_jira.ProjectTemplate'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.7 on 2018-07-06 15:14
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('structure', '0002_immutable_default_json'),
        ('marketplace', '0012_offering_default_value'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='offering',
            name='provider',
        ),
        migrations.AddField(
            model_name='offering',
            name='customer',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='+', to='structure.Customer'),
        ),
    ]



from celery import chain

from waldur_core.core import tasks, executors


class ProjectCreateExecutor(executors.CreateExecutor):

    @classmethod
    def get_task_signature(cls, project, serialized_project, **kwargs):
        return tasks.BackendMethodTask().si(
            serialized_project, 'create_project', state_transition='begin_creating')


class ProjectUpdateExecutor(executors.UpdateExecutor):

    @classmethod
    def get_task_signature(cls, project, serialized_project, **kwargs):
        return tasks.BackendMethodTask().si(
            serialized_project, 'update_project', state_transition='begin_updating')


class ProjectImportExecutor(executors.UpdateExecutor):

    @classmethod
    def get_task_signature(cls, project, serialized_project, **kwargs):
        return tasks.BackendMethodTask().si(
            serialized_project, 'import_project_issues', state_transition='begin_updating')


class ProjectDeleteExecutor(executors.DeleteExecutor):

    @classmethod
    def get_task_signature(cls, project, serialized_project, **kwargs):
        if project.backend_id:
            return tasks.BackendMethodTask().si(
                serialized_project, 'delete_project', state_transition='begin_deleting')
        else:
            return tasks.StateTransitionTask().si(serialized_project, state_transition='begin_deleting')


class IssueCreateExecutor(executors.CreateExecutor):

    @classmethod
    def get_task_signature(cls, issue, serialized_issue, **kwargs):
        return tasks.BackendMethodTask().si(
            serialized_issue, 'create_issue', state_transition='begin_creating')


class IssueUpdateExecutor(executors.UpdateExecutor):

    @classmethod
    def get_task_signature(cls, issue, serialized_issue, **kwargs):
        return tasks.BackendMethodTask().si(
            serialized_issue, 'update_issue', state_transition='begin_updating')


class IssueUpdateFromBackendExecutor(executors.ActionExecutor):

    @classmethod
    def get_task_signature(cls, issue, serialized_issue, **kwargs):
        return tasks.BackendMethodTask().si(
            serialized_issue, 'update_issue_from_jira', state_transition='begin_updating')


class IssueDeleteExecutor(executors.DeleteExecutor):

    @classmethod
    def get_task_signature(cls, issue, serialized_issue, **kwargs):
        if issue.backend_id:
            return tasks.BackendMethodTask().si(
                serialized_issue, 'delete_issue', state_transition='begin_deleting')
        else:
            return tasks.StateTransitionTask().si(serialized_issue, state_transition='begin_deleting')


class CommentCreateExecutor(executors.CreateExecutor):

    @classmethod
    def get_task_signature(cls, comment, serialized_comment, **kwargs):
        return tasks.BackendMethodTask().si(
            serialized_comment, 'create_comment', state_transition='begin_creating')


class CommentUpdateExecutor(executors.UpdateExecutor):

    @classmethod
    def get_task_signature(cls, comment, serialized_comment, **kwargs):
        return tasks.BackendMethodTask().si(
            serialized_comment, 'update_comment', state_transition='begin_updating')


class CommentDeleteExecutor(executors.DeleteExecutor):

    @classmethod
    def get_task_signature(cls, comment, serialized_comment, **kwargs):
        if comment.backend_id:
            return tasks.BackendMethodTask().si(
                serialized_comment, 'delete_comment', state_transition='begin_deleting')
        else:
            return tasks.StateTransitionTask().si(serialized_comment, state_transition='begin_deleting')


class AttachmentCreateExecutor(executors.CreateExecutor):

    @classmethod
    def get_task_signature(cls, attachment, serialized_attachment, **kwargs):
        return tasks.BackendMethodTask().si(
            serialized_attachment, 'create_attachment', state_transition='begin_creating')


class AttachmentDeleteExecutor(executors.DeleteExecutor):

    @classmethod
    def get_task_signature(cls, attachment, serialized_attachment, **kwargs):
        if attachment.backend_id:
            return tasks.BackendMethodTask().si(
                serialized_attachment, 'delete_attachment', state_transition='begin_deleting')
        else:
            return tasks.StateTransitionTask().si(serialized_attachment, state_transition='begin_deleting')


class ProjectPullExecutor(executors.ActionExecutor):
    action = 'Synchronize'

    @classmethod
    def get_action_details(cls, project, **kwargs):
        if 'issues_count' not in project.action_details:
            backend = project.get_backend()
            issues_count = backend.get_issues_count(project.backend_id)
            return {'issues_count': issues_count,
                    'current_issue': 0,
                    'percentage': 0}

    @classmethod
    def get_task_signature(cls, project, serialized_project, **kwargs):
        return chain(
            tasks.StateTransitionTask().si(
                serialized_project,
                state_transition='begin_updating'
            ),
            tasks.PollRuntimeStateTask().si(
                serialized_project,
                backend_pull_method='import_project_batch',
                success_state='success',
                erred_state='error',
            )
        )



import logging

from django.core.management.base import BaseCommand
from rest_framework import status

from waldur_core.core.models import User
from waldur_core.logging.views import EventViewSet
from waldur_mastermind.common.utils import get_request
from waldur_mastermind.marketplace.models import Order, OrderItem, Resource


logger = logging.getLogger(__name__)


def get_resource_user(default_user, resource):
    if not resource.scope:
        logger.warning('Using default user for resource because it does not have scope. '
                       'Resource UUID: %s', resource.uuid)
        return default_user
    view = EventViewSet.as_view({'get': 'list'})
    response = get_request(view, default_user,
                           event_type='resource_creation_scheduled',
                           resource_type=resource.scope.get_scope_type(),
                           resource_uuid=resource.scope.uuid)
    if response.status_code == status.HTTP_200_OK and len(response.data) > 0:
        user_uuid = response.data[0]['user_uuid']
        try:
            return User.objects.get(uuid=user_uuid)
        except User.DoesNotExist:
            logger.warning('Using default user for resource because user does not exist. '
                           'Resource UUID: %s. User UUID: %s', resource.uuid, user_uuid)
    else:
        logger.warning('Using default user for resource because related event is not found. '
                       'Resource UUID: %s', resource.uuid)
    return default_user


def import_orders(find_user=True):
    default_user = User.objects.filter(is_staff=True).first()
    existing_resources = OrderItem.objects.exclude(resource_id=None) \
        .values_list('resource_id', flat=True).distinct()
    missing_resources = Resource.objects.exclude(id__in=existing_resources)
    for resource in missing_resources:
        user = find_user and get_resource_user(default_user, resource) or default_user
        order = Order.objects.create(
            created=resource.created,
            modified=resource.modified,
            created_by=user,
            approved_by=user,
            approved_at=resource.created,
            project=resource.project,
            state=Order.States.DONE,
        )
        OrderItem.objects.create(
            order=order,
            resource=resource,
            offering=resource.offering,
            attributes=resource.attributes,
            limits=resource.limits,
            plan=resource.plan,
            state=OrderItem.States.DONE,
        )
    return missing_resources.count()


class Command(BaseCommand):
    help = """Create marketplace order for each resource if it does not yet exist."""

    def handle(self, *args, **options):
        count = import_orders()
        if count == 0:
            self.stdout.write(self.style.SUCCESS('There are no resources without orders.'))
        if count == 1:
            self.stdout.write(self.style.SUCCESS('%s order has been created.' % count))
        else:
            self.stdout.write(self.style.SUCCESS('%s orders have been created.' % count))



# -*- coding: utf-8 -*-
# Generated by Django 1.11.7 on 2018-07-03 17:43
from __future__ import unicode_literals

from django.db import migrations
import waldur_core.core.fields


class Migration(migrations.Migration):

    dependencies = [
        ('waldur_azure', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='virtualmachine',
            name='private_ips',
            field=waldur_core.core.fields.JSONField(blank=True, default=list, help_text='List of private IP addresses'),
        ),
        migrations.AlterField(
            model_name='virtualmachine',
            name='public_ips',
            field=waldur_core.core.fields.JSONField(blank=True, default=list, help_text='List of public IP addresses'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.18 on 2019-02-01 12:03
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('waldur_azure', '0008_networkinterface_public_ip'),
    ]

    operations = [
        migrations.AddField(
            model_name='publicip',
            name='resource_group',
            field=models.ForeignKey(default=None, on_delete=django.db.models.deletion.CASCADE, to='waldur_azure.ResourceGroup'),
            preserve_default=False,
        ),
    ]



import string

import requests

from . import models


def find_versions(queried_library_name, python_version):
    lib_info = requests.get("https://pypi.python.org/pypi/%s/json" % queried_library_name).json()

    wrong_os_versions = ["win", "mac"]
    python_versions = build_version_criteria(python_version)
    version_upload_date_pairs = []

    for release_version, release_info in lib_info['releases'].viewitems():
        for platform_version in release_info:
            if not contains_any_string(platform_version['filename'], wrong_os_versions) and\
                    (contains_any_string(platform_version['python_version'], python_versions) or
                     version_not_specified(platform_version['filename'])):
                version_upload_date_pairs.append(dict(version=release_version, upload_date=platform_version['upload_time']))
                break

    version_upload_date_pairs.sort(reverse=True, key=lambda pair: pair['upload_date'])
    return map(lambda pair: pair['version'], version_upload_date_pairs)


def build_version_criteria(python_version):
    if python_version != '3':
        required_version = python_version[0:find_nth(python_version, '.', 2)]
        required_version_without_dots = string.replace(required_version, '.', '').strip()
        return [required_version, required_version_without_dots]
    else:
        return [python_version]


def find_nth(target_string, substring, n):
    start = target_string.find(substring)
    while start >= 0 and n > 1:
        start = target_string.find(substring, start + len(substring))
        n -= 1
    return start


def version_not_specified(python_version):
    return 'cp' not in python_version and 'py' not in python_version


def contains_any_string(target_string, strings):
    for item in strings:
        if item in target_string:
            return True
    return False


def autocomplete_library_name(queried_library_name):
    return models.CachedRepositoryPythonLibrary.objects.filter(name__startswith=queried_library_name).order_by('name')[0:30]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models
import django.utils.timezone
from django.conf import settings
import model_utils.fields
import waldur_core.core.fields


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('logging', '0005_add_alert_unique_together_constraint'),
    ]

    operations = [
        migrations.CreateModel(
            name='PushHook',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, verbose_name='created', editable=False)),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, verbose_name='modified', editable=False)),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('event_types', waldur_core.core.fields.JSONField(verbose_name='List of event types')),
                ('is_active', models.BooleanField(default=True)),
                ('last_published', models.DateTimeField(default=django.utils.timezone.now)),
                ('type', models.SmallIntegerField(choices=[(1, 'iOS'), (2, 'Android')])),
                ('registration_token', models.CharField(max_length=255, blank=True)),
                ('user', models.ForeignKey(to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='SystemNotification',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('event_types', waldur_core.core.fields.JSONField(verbose_name='List of event types')),
                ('hook_content_type', models.OneToOneField(related_name='+', to='contenttypes.ContentType')),
            ],
        ),
    ]



from datetime import timedelta
import logging

from celery import shared_task

from django.contrib.auth import get_user_model
from django.utils import timezone

from waldur_core.core import tasks
from waldur_core.core import views

from . import client, models


logger = logging.getLogger(__name__)


class AuthTask(tasks.StateTransitionTask):
    """ Execute request for authentication """

    def execute(self, auth_result):
        response = client.SignatureRequest.execute(
            transaction_id=auth_result.uuid.hex,
            phone=auth_result.phone,
            message=auth_result.message)
        auth_result.backend_transaction_id = response.backend_transaction_id
        auth_result.save(update_fields=['backend_transaction_id'])


class PollTask(views.RefreshTokenMixin, tasks.Task):
    max_retries = 25
    default_retry_delay = 12

    def execute(self, auth_result):
        response = client.StatusRequest.execute(
            transaction_id=auth_result.uuid.hex,
            backend_transaction_id=auth_result.backend_transaction_id)
        if response.status == client.Statuses.OK:
            self._associate_with_user(auth_result, response.civil_number)
        elif response.status == client.Statuses.PROCESSING:
            self.retry()
        elif response.status == client.Statuses.ERRED:
            auth_result.set_canceled()
            auth_result.details = response.details
            auth_result.save(update_fields=['state', 'details'])
            logger.info('PKI login failed for user with phone %s, details: %s.', auth_result.phone, auth_result.details)

    def _associate_with_user(self, auth_result, civil_number):
        User = get_user_model()
        try:
            user = User.objects.get(civil_number=civil_number)
            self.refresh_token(user)
            auth_result.user = user
            auth_result.set_ok()
            logger.info('PKI login was successfully done for user %s.', user.username)
        except User.DoesNotExist:
            auth_result.details = 'User is not registered.'
            auth_result.set_canceled()
            logger.info('PKI login failed for user with civil number %s - user record does not exist in Waldur.', civil_number)
        auth_result.save()


@shared_task(name='waldur_auth_valimo.cleanup_auth_results')
def cleanup_auth_results():
    models.AuthResult.objects.filter(modified__lte=timezone.now() - timedelta(days=7)).delete()



# -*- coding: utf-8 -*-
# Generated by Django 1.11.13 on 2018-08-31 11:16
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('marketplace', '0029_orderitem_state'),
    ]

    operations = [
        migrations.AddField(
            model_name='attribute',
            name='required',
            field=models.BooleanField(default=False, help_text='A value must be provided for the attribute.'),
        ),
    ]



from django.db import transaction
from django.utils.translation import ugettext_lazy as _
from rest_framework import decorators, viewsets, \
    response, status, serializers as rf_serializers

from waldur_core.core import validators as core_validators
from waldur_core.structure import views as structure_views

from . import models, serializers, executors, filters


class AzureServiceViewSet(structure_views.BaseServiceViewSet):
    queryset = models.AzureService.objects.all()
    serializer_class = serializers.ServiceSerializer


class AzureServiceProjectLinkViewSet(structure_views.BaseServiceProjectLinkViewSet):
    queryset = models.AzureServiceProjectLink.objects.all()
    serializer_class = serializers.ServiceProjectLinkSerializer


class ImageViewSet(structure_views.BaseServicePropertyViewSet):
    queryset = models.Image.objects.all()
    serializer_class = serializers.ImageSerializer
    filter_class = filters.ImageFilter
    lookup_field = 'uuid'

    def get_queryset(self):
        return models.Image.objects.order_by('name')


class SizeViewSet(viewsets.ReadOnlyModelViewSet):
    queryset = models.Size.objects.all()
    serializer_class = serializers.SizeSerializer
    filter_class = filters.SizeFilter
    lookup_field = 'uuid'


class LocationViewSet(viewsets.ReadOnlyModelViewSet):
    queryset = models.Location.objects.filter(enabled=True)
    serializer_class = serializers.LocationSerializer
    filter_class = filters.LocationFilter
    lookup_field = 'uuid'


class ResourceGroupViewSet(viewsets.ReadOnlyModelViewSet):
    queryset = models.ResourceGroup.objects.all()
    serializer_class = serializers.ResourceGroupSerializer
    lookup_field = 'uuid'


class PublicIPViewSet(structure_views.BaseResourceViewSet):
    queryset = models.PublicIP.objects.all()
    filter_class = filters.PublicIPFilter
    serializer_class = serializers.PublicIPSerializer
    create_executor = executors.PublicIPCreateExecutor
    delete_executor = executors.PublicIPDeleteExecutor


class VirtualMachineViewSet(structure_views.BaseResourceViewSet):
    queryset = models.VirtualMachine.objects.all()
    filter_class = filters.VirtualMachineFilter
    serializer_class = serializers.VirtualMachineSerializer
    create_executor = executors.VirtualMachineCreateExecutor
    delete_executor = executors.VirtualMachineDeleteExecutor

    @decorators.detail_route(methods=['post'])
    def start(self, request, uuid=None):
        virtual_machine = self.get_object()
        executors.VirtualMachineStartExecutor().execute(virtual_machine)
        return response.Response({'status': _('start was scheduled')}, status=status.HTTP_202_ACCEPTED)

    start_validators = [core_validators.StateValidator(models.VirtualMachine.States.OK),
                        core_validators.RuntimeStateValidator('stopped')]
    start_serializer_class = rf_serializers.Serializer

    @decorators.detail_route(methods=['post'])
    def stop(self, request, uuid=None):
        virtual_machine = self.get_object()
        executors.VirtualMachineStopExecutor().execute(virtual_machine)
        return response.Response({'status': _('stop was scheduled')}, status=status.HTTP_202_ACCEPTED)

    stop_validators = [core_validators.StateValidator(models.VirtualMachine.States.OK),
                       core_validators.RuntimeStateValidator('running')]
    stop_serializer_class = rf_serializers.Serializer

    @decorators.detail_route(methods=['post'])
    def restart(self, request, uuid=None):
        virtual_machine = self.get_object()
        executors.VirtualMachineRestartExecutor().execute(virtual_machine)
        return response.Response({'status': _('restart was scheduled')}, status=status.HTTP_202_ACCEPTED)

    restart_validators = [core_validators.StateValidator(models.VirtualMachine.States.OK),
                          core_validators.RuntimeStateValidator('running')]
    restart_serializer_class = rf_serializers.Serializer


class SQLServerViewSet(structure_views.BaseResourceViewSet):
    queryset = models.SQLServer.objects.all()
    filter_class = filters.SQLServerFilter
    serializer_class = serializers.SQLServerSerializer
    create_executor = executors.SQLServerCreateExecutor
    delete_executor = executors.SQLServerDeleteExecutor

    @decorators.detail_route(methods=['post'])
    def create_database(self, request, uuid=None):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        database = serializer.save()

        transaction.on_commit(lambda: executors.SQLDatabaseCreateExecutor().execute(database))

        payload = {
            'status': _('SQL database creation was scheduled'),
            'database_uuid': database.uuid,
        }
        return response.Response(payload, status=status.HTTP_202_ACCEPTED)

    create_database_validators = [core_validators.StateValidator(models.SQLServer.States.OK)]
    create_database_serializer_class = serializers.SQLDatabaseCreateSerializer


class SQLDatabaseViewSet(structure_views.BaseResourceViewSet):
    queryset = models.SQLDatabase.objects.all()
    filter_class = filters.SQLDatabaseFilter
    serializer_class = serializers.SQLDatabaseSerializer
    create_executor = executors.SQLDatabaseCreateExecutor
    delete_executor = executors.SQLDatabaseDeleteExecutor



# -*- coding: utf-8 -*-
# Generated by Django 1.11.18 on 2019-04-02 07:58
from __future__ import unicode_literals

import django.contrib.postgres.fields.jsonb
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('marketplace', '0073_plan_period_to_usage'),
    ]

    operations = [
        migrations.AddField(
            model_name='attribute',
            name='default',
            field=django.contrib.postgres.fields.jsonb.JSONField(blank=True, null=True),
        ),
    ]
